'''
'''

from __future__ import absolute_import

import os
from base64 import b64decode

from cottoncandy import options

from .browser import BrowserObject
from .interfaces import InterfaceObject
from .utils import get_keys, string2bool

__version__ = "0.3.0"

ACCESS_KEY = options.config.get('login', 'access_key')
SECRET_KEY = options.config.get('login', 'secret_key')
ENDPOINT_URL = options.config.get('login', 'endpoint_url')
DEFAULT_SIGNATURE_VERSION = options.config.get('basic', 'signature_version')

default_bucket = options.config.get('basic', 'default_bucket')
force_bucket_creation = options.config.get('basic', 'force_bucket_creation')
force_bucket_creation = string2bool(force_bucket_creation)

def get_interface(bucket_name=default_bucket,
                  ACCESS_KEY=ACCESS_KEY,
                  SECRET_KEY=SECRET_KEY,
                  endpoint_url=ENDPOINT_URL,
                  force_bucket_creation=force_bucket_creation,
                  verbose=True,
                  backend='s3',
                  **kwargs):
    """Return an interface to the cloud.

    Parameters
    ----------
    bucket_name : str
    ACCESS_KEY : str
    SECRET_KEY : str
    endpoint_url : str
        The URL for the S3 gateway
    backend : 's3'|'gdrive'
        What backend to hook on to
    kwargs :
        S3 only. kwargs passed to botocore. For example,
        >>> from botocore.client import Config
        >>> config = Config(connect_timeout=50, read_timeout=10*60)
        >>> cci = cc.get_interface('my_bucket', config=config)

    Returns
    -------
    cci : cottoncandy.InterfaceObject
    """
    from cottoncandy.interfaces import DefaultInterface

    if backend == 's3':
        if ACCESS_KEY in [False, "False"] or SECRET_KEY in [False, "False"]:
            ACCESS_KEY, SECRET_KEY = get_keys()
    elif backend == 'gdrive':
        ACCESS_KEY = os.path.join(options.userdir, options.config.get('gdrive', 'secrets'))
        SECRET_KEY = os.path.join(options.userdir, options.config.get('gdrive', 'credentials'))
    else:
        pass

    if 'config' in kwargs:
        # user provided config
        if not kwargs['config'].signature_version:
            # config does not specify signature
            kwargs['config'].signature_version = DEFAULT_SIGNATURE_VERSION
    elif DEFAULT_SIGNATURE_VERSION:
        # no config but default signature exists
        from botocore.client import Config
        kwargs['config'] = Config(signature_version=DEFAULT_SIGNATURE_VERSION)

    interface = DefaultInterface(bucket_name,
                                 ACCESS_KEY,
                                 SECRET_KEY,
                                 endpoint_url,
                                 force_bucket_creation,
                                 verbose=verbose,
                                 backend = backend,
                                 **kwargs)
    return interface


def get_browser(bucket_name=default_bucket,
                ACCESS_KEY=ACCESS_KEY,
                SECRET_KEY=SECRET_KEY,
                endpoint_url=ENDPOINT_URL):
    """Browser object that allows you to tab-complete your
    way through your objects

    Parameters
    ----------
    bucket_name : str
    ACCESS_KEY : str
    SECRET_KEY : str
    endpoint_url : str
        The URL for the S3 gateway

    Returns
    -------
    ccb : cottoncandy.BrowserObject

    Example
    -------
    >>> browser = cc.get_browser('my_bucket',
                                 ACCESS_KEY='FAKEACCESSKEYTEXT',
                                 SECRET_KEY='FAKESECRETKEYTEXT',
                                 endpoint_url='https://s3.amazonaws.com')
    >>> browser.sweet_project.sub<TAB>
    browser.sweet_project.sub01_awesome_analysis_DOT_grp
    browser.sweet_project.sub02_awesome_analysis_DOT_grp
    >>> browser.sweet_project.sub01_awesome_analysis_DOT_grp
    <cottoncandy-group <bucket:my_bucket_name> (sub01_awesome_analysis.grp: 3 keys)>
    >>> browser.sweet_project.sub01_awesome_analysis_DOT_grp.result_model01
    <cottoncandy-dataset <bucket:my_bucket_name [1.00MB:shape=(10000)]>
    """
    from cottoncandy.browser import S3Directory
    from cottoncandy.interfaces import DefaultInterface

    if (ACCESS_KEY is False) and (SECRET_KEY is False):
        from .utils import get_keys
        ACCESS_KEY, SECRET_KEY = get_keys()

    interface = DefaultInterface(bucket_name,
                                 ACCESS_KEY,
                                 SECRET_KEY,
                                 endpoint_url,
                                 force_bucket_creation=False,
                                 verbose=False)

    return S3Directory('/', interface=interface)

__all__ = ['get_interface', 'get_browser', 'interfaces', 'browser']



import os
import json
import glob
import shutil

try:
    from cStringIO import StringIO
except ImportError:
    from io import BytesIO as StringIO

from .backend import CCBackEnd, CloudStream
from .utils import sanitize_metadata
from .utils import remove_root
from .utils import remove_trivial_magic
from .utils import SEPARATOR

METADATA_SUFFIX = ".meta.json"


class LocalClient(CCBackEnd):
    """
    Client interface for local file system.

    Handle metadata in CouldStream objects by storing a json file (.meta.json).
    """

    def __init__(self, path):
        if not os.path.isdir(path):
            os.makedirs(path)
        self.path = path

    def check_file_exists(self, cloud_name, bucket_name=None):
        """Checks whether a file exists on the cloud

        Parameters
        ----------
        cloud_name : str
            file on cloud
        bucket_name : str
            For a local client, the bucket is just a directory path.
            If None, use self.path.

        Returns
        -------
        bool
        """
        if bucket_name is None:
            bucket_name = self.path
        return os.path.isfile(os.path.join(bucket_name, cloud_name))

    def upload_stream(self, stream, cloud_name, metadata, permissions):
        """Uploads a stream object with a .read() function

        Parameters
        ----------
        stream : stream
            streaming object, e.g. StringIO(json.dumps(my_dict)) (no metadata)
        cloud_name : str
            name to use on cloud
        metadata : dict
            custom metadata for this file
        permissions : ignored

        Returns
        -------
        bool, upload success
        """
        file_name = os.path.join(self.path, cloud_name)
        auto_makedirs(file_name)
        with open(file_name, 'wb') as local_file:
            local_file.write(stream.read())

        metadata_file_name = file_name + METADATA_SUFFIX
        with open(metadata_file_name, 'w') as local_file:
            json.dump(metadata, local_file, indent=4)

    def upload_file(self, file_name, cloud_name, permissions):
        """Uploads a file from disk

        Parameters
        ----------
        file_name : str
            name of file to upload
        cloud_name : str
            name to use on the cloud
        permissions : ignored

        Returns
        -------
        bool, upload success
        """
        destination = os.path.join(self.path, cloud_name)
        auto_makedirs(destination)
        self.copy(
            source=file_name,
            destination=destination,
            source_bucket=None,
            destination_bucket=None,
            overwrite=True,
        )

    def upload_multipart(self, stream, cloud_name, metadata, permissions,
                         buffersize, verbose):
        """Multi-part upload for large stream objects

        Parameters
        ----------
        stream : stream
            streaming object
        cloud_name : str
            name to use on cloud
        metadata : dict
            custom metadata
        permissions : str?
            permissions for this file
        buffersize : int
            s3 uploading buffersize
        verbose : bool
            s3 verbosity

        Returns
        -------
        bool, upload success
        """
        return self.upload_stream(stream, cloud_name, metadata, permissions)

    def download_stream(self, cloud_name):
        """Downloads a object to an in-memory stream

        Parameters
        ----------
        cloud_name : str
            name of object to download

        Returns
        -------
        CloudStream object
        """
        file_name = os.path.join(self.path, cloud_name)
        with open(file_name, 'rb') as local_file:
            content = StringIO()
            content.write(local_file.read())  # load in memory
            content.seek(0)

        metadata_file_name = file_name + METADATA_SUFFIX
        if os.path.isfile(metadata_file_name):
            with open(metadata_file_name, 'r') as local_file:
                metadata = json.load(local_file)
        else:
            metadata = dict()

        return CloudStream(content, sanitize_metadata(metadata))

    def download_to_file(self, cloud_name, file_name):
        """Downloads an object directly to disk

        Parameters
        ----------
        cloud_name : str
            name of object to download
        file_name : str
            name on disk to use

        Returns
        -------
        bool, download success
        """
        return self.copy(
            source=os.path.join(self.path, cloud_name),
            destination=file_name,
            source_bucket=None,
            destination_bucket=None,
            overwrite=True,
        )

    def list_directory(self, path, limit):
        """Lists the content of a directory

        Parameters
        ----------
        path : str
            path on the cloud to get contents for
        limit : ignored

        Returns
        -------

        """
        if (path != '') and (path != '/'):
            path = remove_root(path)
        path = remove_trivial_magic(path)

        path = os.path.join(self.path, path)
        results = glob.glob(os.path.join(path, "*"))
        results += glob.glob(os.path.join(path, ".*"))  # hidden files
        results = self._remove_path_and_metadata(results)
        return results

    def list_objects(self, **kwargs):
        """Gets all objects contained by backend

        Returns
        -------

        """
        # match S3Client API to get a prefix
        filter = kwargs.pop("filter", dict())
        prefix = filter.pop("Prefix", "")

        results = glob.glob(os.path.join(self.path, prefix, "**", "*"),
                            recursive=True)
        results += glob.glob(os.path.join(self.path, prefix, "**", ".*"),
                             recursive=True)  # hidden files
        results += glob.glob(os.path.join(self.path, prefix, "**", ".*", "**"),
                             recursive=True)  # files in hidden directories
        # remove directories
        results = [res for res in results if os.path.isfile(res)]
        results = self._remove_path_and_metadata(results)
        return results

    def copy(self, source, destination, source_bucket, destination_bucket,
             overwrite):
        """Copies an object

        Parameters
        ----------
        source : str
            origin path
        destination : str
            destination path
        source_bucket : ignored
        destination_bucket : ignored
        overwrite : bool
            overwrite if destination exists

        Returns
        -------
        bool, copy success
        """
        if overwrite is False:
            raise NotImplementedError()
        if source_bucket is None:
            source_bucket = self.path
        if destination_bucket is None:
            destination_bucket = source_bucket
        source = os.path.join(source_bucket, source)
        destination = os.path.join(destination_bucket, destination)
        auto_makedirs(destination)
        return shutil.copy(source, destination)

    def move(self, source, destination, source_bucket, destination_bucket,
             overwrite):
        """Moves an object

        Parameters
        ----------
        source
        destination
        source_bucket
        destination_bucket
        overwrite

        Returns
        -------

        """
        if overwrite is False:
            raise NotImplementedError()
        if source_bucket is None:
            source_bucket = self.path
        if destination_bucket is None:
            destination_bucket = source_bucket
        source = os.path.join(source_bucket, source)
        destination = os.path.join(destination_bucket, destination)
        auto_makedirs(destination)
        return shutil.move(source, destination)

    def delete(self, cloud_name, recursive=False, delete=False):
        """Deletes an object

        Parameters
        ----------
        cloud_name : str
            name of cloud object to delete
        recursive : bool
            recursively delete directory?
        delete : bool
            ignored

        Returns
        -------

        """
        cloud_name = os.path.join(self.path, cloud_name)
        if os.path.isfile(cloud_name):
            os.remove(cloud_name)
        else:
            if recursive:
                shutil.rmtree(cloud_name)
            else:
                os.rmdir(cloud_name)

    @property
    def size(self):
        """Size of stored cloud items in bytes

        Returns
        -------
        int
        """
        total_size = 0
        for dirpath, dirnames, filenames in os.walk(self.path):
            for f in filenames:
                fp = os.path.join(dirpath, f)
                # skip if it is symbolic link
                if not os.path.islink(fp):
                    total_size += os.path.getsize(fp)

        return total_size

    def _remove_path_and_metadata(self, file_list, path=None):
        """Removes path from filenames, removes .meta.json files from the list.
        """
        if path is None:
            path = self.path
        results = []
        for file_name in file_list:
            # remove path
            if file_name.startswith(path):
                file_name = file_name[len(path):]
            if file_name[0] == SEPARATOR:
                file_name = file_name[1:]
            # remove .meta.json files
            if file_name.endswith(METADATA_SUFFIX):
                continue
            else:
                results.append(file_name)
        return results

    def get_object_metadata(self, object_name):
        """Get metadata associated with an object"""
        file_name = os.path.join(self.path, object_name)

        metadata_file_name = file_name + METADATA_SUFFIX
        if os.path.isfile(metadata_file_name):
            with open(metadata_file_name, 'r') as local_file:
                try:
                    metadata = json.load(local_file)
                except Exception as e:
                    print(local_file)
                    raise e
        else:
            metadata = dict()
        metadata = sanitize_metadata(metadata)

        return metadata

    def get_object_size(self, object_name):
        """Get the size in bytes of an object"""
        file_name = os.path.join(self.path, object_name)
        size = os.path.getsize(file_name)
        return size


def auto_makedirs(destination):
    """Create directory tree if destination does not exist."""
    if not os.path.exists(os.path.dirname(destination)):
        os.makedirs(os.path.dirname(destination))



'''
'''
import os
import re

from cottoncandy.utils import (clean_object_name,
                               has_start_digit,
                               has_magic,
                               has_real_magic,
                               has_trivial_magic,
                               remove_trivial_magic,
                               remove_root,
                               mk_aws_path,
                               objects2names,
                               unquote_names,
                               print_objects,
                               get_fileobject_size,
                               get_object_size,
                               read_buffered,
                               GzipInputStream,
                               generate_ndarray_chunks,
                               sanitize_metadata,
                               )




# globals
##############################

HDFEXT = ['.hdf', '.h5', '.hf5', '.hdf5', '.h5py', '.grp', '.arr', '.dar']
NPYEXT = ['.npy']
JSONEXT = ['.json']


#------------------
# Browsing objects
#------------------

class BrowserObject(object):
    pass

class S3FSLike(BrowserObject):
    '''Base class for file-system-like backend_interface to S3
    '''
    @clean_object_name
    def __init__(self, path, interface):
        '''
        path: str
            The path to start naviation from
        backend_interface:
            A `cottoncandy.InterfaceObject` instance
        '''

        self.interface = interface
        parent, curdir = os.path.split(path)
        self._curdir = curdir
        self._parent = parent

    @property
    def _fullpath(self):
        return os.path.join(self._parent, self._curdir)

    def __call__(self):
        assert self.interface.exists_object(self._fullpath)
        return self.interface.get_object(self._fullpath)


class S3Directory(S3FSLike):
    """Get an object that allows you to tab-complete your
    way through your objects.
    One also has access to ``_ls()`` and ``_fullpath()`` methods.
    """
    @clean_object_name
    def __init__(self, path, interface):
        """
        Parameters
        ----------
        path: str
            The path to start naviation from
        interface:
            A `cottoncandy.InterfaceObject` instance

        Returns
        -------
        ccbrowser : cottoncandy.BrowserObject
            A cottoncandy browser object

        Examples
        --------
        >>> import cottoncandy as cc
        >>> browser = cc.get_browser('my_bucket_name')
        >>> browser.s<TAB>
        browser.sweet_project
        >>> browser.sweet_project
        cottoncandy-path <bucket:my_bucket_name> sweet_project

        >>> browser.sweet_project.sub<TAB>
        browser.sweet_project.sub01_awesome_analysis_DOT_grp
        browser.sweet_project.sub02_awesome_analysis_DOT_grp
        >>> browser.sweet_project.sub01_awesome_analysis_DOT_grp
        <cottoncandy-group <bucket:my_bucket_name> (sub01_awesome_analysis.grp: 3 keys)>

        We can also explore the contents of the ``grp`` object

        >>> browser.sweet_project.sub01_awesome_analysis_DOT_grp.<TAB>
        browser.sweet_project.sub01_awesome_analysis_DOT_grp.result_model01
        browser.sweet_project.sub01_awesome_analysis_DOT_grp.result_model02
        browser.sweet_project.sub01_awesome_analysis_DOT_grp.result_model03

        Let's look at one

        >>> browser.sweet_project.sub01_awesome_analysis_DOT_grp.result_model01
        <cottoncandy-dataset <bucket:my_bucket_name [1.00MB:shape=(10000)]>

        We can download the data

        >>> arr = browser.sweet_project.sub01_awesome_analysis_DOT_grp.result_model01.load()
        >>> arr.shape
        (10000,)

        We can also ask for the object name in the bucket

        >>> browser.sweet_project.sub01_awesome_analysis_DOT_grp._fullpath
        'sweet_project/sub01_awesome_analysis.grp/result_model01'

        We can also download the object. Useful when dealing with other object types

        >>> browser.sweet_project.sub01_awesome_analysis_DOT_grp.result_model01()
        s3.Object(bucket_name='my_bucket_name', key='sweet_project/sub01_awesome_analysis.grp/result_model01')




        """
        super(S3Directory, self).__init__(path, interface=interface)

        subdirs = self.interface.lsdir(self._fullpath)
        subdirs = [os.path.split(t)[-1] for t in subdirs]
        self._subdirs = {}

        for sdir in subdirs:
            if sdir:
                # modify name for tab-completion purposes
                sdir_copy = sdir[:]
                # clean numbers
                sdir_copy = 'NUM_%s'%sdir_copy if has_start_digit(sdir) else sdir_copy
                # clean extension
                fl, ext = os.path.splitext(sdir_copy)
                kk = fl+'_DOT_'+ext[1:] if ext else sdir_copy
                # clean fucking dashes
                kk = re.sub('-', '_', kk) if '-' in kk else kk
                self._subdirs[kk] = sdir

    def _ls(self):
        '''Show the directories and files contained at this level
        '''
        return sorted(self._subdirs.keys())

    def _glob(self):
        '''Show all the objects contained under this branch
        '''
        return sorted(self.interface.glob(self._fullpath))

    def __dir__(self):
        return []

    def __repr__(self):
        if len(self._subdirs):
            details = (__package__, self.interface.bucket_name, self._curdir)
            return "%s-path <bucket:%s> %s"%details
        else:
            # no children, it's gotta be an object b/c we're in S3
            obj = self.interface.get_object(self._fullpath)
            size = get_object_size(obj)
            details = (__package__, self.interface.bucket_name, size)
            return "%s-file <bucket:%s> [%0.01fMB]"%details

    def __len__(self):
        return len(self._subdirs)

    def __dir__(self):
        return list(self._subdirs.keys())

    def __getattr__(self, attr):
        if attr in self._subdirs:
            child_path = os.path.join(self._fullpath, self._subdirs[attr])
        else:
            # allow user-provided paths. useful when >1000 objects
            child_path = os.path.join(self._fullpath, attr)
            assert (self.interface.exists_object(child_path) or \
                    self.interface.lsdir(child_path))

        try:
            has_ext = '.' in child_path # shitty check...
            ishdf = any([t in child_path for t in HDFEXT])

            if ishdf:
                return S3HDF5(child_path, interface=self.interface)
            else:
                return S3Directory(child_path, interface=self.interface)

        except:
            raise AttributeError


class S3HDF5(S3Directory):
    '''
    '''
    @clean_object_name
    def __init__(self, path, interface):
        '''
        '''
        super(S3HDF5, self).__init__(path, interface=interface)

    def __repr__(self):
        if len(self._subdirs) == 0:
            # bottom object, probably array
            obj = self.interface.get_object(self._fullpath)
            obmeta = sanitize_metadata(obj.metadata)
            if 'shape' in obmeta:
                shape =  obmeta['shape']
                size = get_object_size(obj)
                details = (__package__, self.interface.bucket_name, size, shape)
                return "%s-dataset <bucket:%s [%0.01fMB:shape=(%s)]>"%details
        # otherwise it's the file itself
        details = (__package__, self.interface.bucket_name,self._curdir, len(self._subdirs))
        return "<%s-group <bucket:%s> (%s: %i keys)>"%details

    def __dir__(self):
        return ['load', 'keys'] + list(self._subdirs.keys())

    def keys(self):
        '''Show the contents of
        '''
        return sorted(self._ls())

    def load(self, key=None):
        '''
        '''
        dataset_path = self._fullpath
        if key is not None:
            if key in self._subdirs:
                dataset_path = os.path.join(self._fullpath, self._subdirs[key])
            else:
                raise ValueError('"%s" not in %s'%(key, dataset_path))

        if self.interface.exists_object(dataset_path):
            return self.interface.download_raw_array(dataset_path)
        print('Specify key to download:\n%s'%','.join(sorted(self._subdirs.keys())))



import os
import sys

from base64 import b64decode, b64encode

try:
    import configparser
except ImportError:
    import ConfigParser as configparser

from . import appdirs

def get_key_from_s3fs():
    '''If user has s3fs-fuse keys,return them
    '''
    key_path = os.path.expanduser('~/.passwd-s3fs')
    if os.path.exists(key_path):
        with open(key_path, 'r') as kfl:
            content = kfl.readline()
            ACCESS_KEY, SECRET_KEY = content.strip().split(':')
            return ACCESS_KEY, SECRET_KEY


def get_key_from_environ():
    try:
        ak = os.environ['AWS_ACCESS_KEY'],
        sk = os.environ['AWS_SECRET_KEY']
        return ak, sk
    except:
        return


def get_keys():
    '''try to find the user keys in the machine
    '''
    # try to outload keys
    resulta = get_key_from_s3fs()
    resultb = get_key_from_environ()
    result = resultb if (resulta is None) else resulta
    return result


def get_config():
    config = configparser.ConfigParser()
    with open(os.path.join(cwd, 'defaults.cfg'), 'r') as defaults_file:
        if sys.version_info.major == 2:
            config.readfp(defaults_file)
        else:
            config.read_file(defaults_file)
    return config

cwd = os.path.split(os.path.abspath(__file__))[0]
userdir = appdirs.user_data_dir("cottoncandy",appauthor="cottoncandy")
usercfg = os.path.join(userdir, "options.cfg")
config = get_config()

# case no user config file
if len(config.read(usercfg)) == 0:
    if not os.path.exists(userdir):
        os.makedirs(userdir)

    prompt = '''
############################################################
Hi! Looks like this is your first time using cottoncandy.

Your cottoncandy configuration file will be stored in:
{path}

You can store your S3 or Google Drive credentials there,
and many other options.

Thanks for using cottoncandy!
'''
    print(prompt.format(path=usercfg))

    with open(usercfg, 'w') as fp:
        config.write(fp)


# add things to old versions of config if needed
else:
    needs_update = False

    try:	# gdrive section
        secrets = config.get('gdrive', 'secrets')
        credentials = config.get('gdrive', 'credentials')
    except configparser.NoSectionError:
        config.add_section('gdrive')
        config.set('gdrive', 'secrets', 'client_secrets.json')
        config.set('gdrive', 'credentials', 'credentials.txt')
        needs_update = True

    if needs_update:
        with open(usercfg, 'w') as configfile:
            config.write(configfile)



import json
import six
import glob

try:
    import cPickle as pickle
except ImportError:
    import pickle

try:
    from urllib import unquote
except ImportError:
    from urllib.parse import unquote

try:
    reduce
except NameError:
    from functools import reduce

import fnmatch
from gzip import GzipFile

try:
    from cStringIO import StringIO
except ImportError:
    from io import BytesIO as StringIO

from base64 import b64decode, b64encode


import cottoncandy.browser
import os
import re

from cottoncandy.backend import FileNotFoundError

from .s3client import S3Client, botocore
from warnings import warn
from .utils import (pathjoin, clean_object_name, print_objects, get_fileobject_size, read_buffered,
                    generate_ndarray_chunks, remove_trivial_magic, has_real_magic, objects2names,
                    has_magic, remove_root, mk_aws_path, string2bool,
                    GzipInputStream,
                    DEFAULT_ACL, MPU_CHUNKSIZE, MPU_THRESHOLD, DASK_CHUNKSIZE, MB, SEPARATOR,
                    MAGIC_CHECK)
from .options import config

DO_COMPRESSION = config.get('compression', 'do_compression').lower() in ('true', 't', 'y','yes')
COMPRESSION_SMALL = config.get('compression', 'small_array')
COMPRESSION_LARGE = config.get('compression', 'large_array')

try:
    import numpy as np
    from scipy.sparse import (coo_matrix,
                            csr_matrix,
                            csc_matrix,
                            bsr_matrix,
                            dia_matrix)
except ImportError:
    warn('numpy/scipy not available')

try:
    import numcodecs
except ImportError:
    warn('numcodecs python library not available')



# ------------------
# Cloud Interfaces
# ------------------

class InterfaceObject(object):
    pass


class BasicInterface(InterfaceObject):
    """Basic cottoncandy interface to the cloud.
    """

    def __init__(self, bucket_name,
                 ACCESS_KEY, SECRET_KEY, url=None,
                 force_bucket_creation=False,
                 verbose=True, backend='s3', **kwargs):
        """
        Parameters
        ----------
        bucket_name : str
        ACCESS_KEY : str
            The S3 access key, or client secrets json file
        SECRET_KEY : str
            The S3 secret key, or client credentials file
        url : str
            The URL for the S3 gateway
        force_bucket_creation: bool
            if bucket does not exist, make it?
        verbose: bool
            print things?
        backend: 's3'|'gdrive'
            Access s3 or google drive?
        kwargs : dict,
            S3 only. Passed to backend.

        Returns
        -------
        cci  : ccio
            Cottoncandy interface object
        """
        bucket_name = None if bucket_name == '' else bucket_name

        if backend == 's3':
            self.backend_interface = S3Client(bucket_name, ACCESS_KEY, SECRET_KEY, url,
                                              force_bucket_creation,
                                              **kwargs)
        elif backend == 'gdrive':
            from .gdriveclient import GDriveClient
            self.backend_interface = GDriveClient(ACCESS_KEY, SECRET_KEY)
        elif backend == 'local':
            from .localclient import LocalClient
            self.backend_interface = LocalClient(path=bucket_name)
        else:
            raise ValueError('Bad backend')

        if verbose:
            if backend == 's3' and bucket_name:
                print('Available buckets:')
                self.show_buckets()
                print('Current bucket: {}'.format(self.backend_interface.bucket_name))
            elif backend == 'gdrive':
                print('Google drive backend instantiated.')
            elif backend == 'local':
                print('Local backend instantiated.')

        self.backend = backend

    def __repr__(self):
        if self.backend == "s3":
            details = (__package__, self.bucket_name, self.backend_interface.url)
            return '%s.backend_interface <bucket:%s on %s>' % details
        elif self.backend == "gdrive":
            return '{}.backend_interface on Google Drive'.format(__package__)
        else:
            details = (__package__, self.backend_interface.path)
            return '%s.backend_interface on local machine (%s)' % details

    def _get_bucket_name(self, bucket_name):
        return self.backend_interface._get_bucket_name(bucket_name)

    def pathjoin(self, a, *p):
        return pathjoin(a, *p)

    @property
    def bucket_name(self):
        if self.backend == "s3":
            return self.backend_interface.bucket_name
        elif self.backend == "gdrive":
            print('Google drive has no concept of buckets')
            return None
        else:
            return self.backend_interface.path

    @clean_object_name
    def exists_object(self, object_name, bucket_name=None, raise_err=False):
        """Check whether object exists in bucket

        Parameters
        ----------
        object_name : str
            The object name
        raise_err : boolean
            If set to True, this function will throw an exception if the
            object does not exist.
        """
        exists = self.backend_interface.check_file_exists(object_name, bucket_name)
        if raise_err and not exists:
            raise FileNotFoundError('Object not found: ' + object_name)
        else:
            return exists

    def exists_bucket(self, bucket_name):
        """Check whether the bucket exists"""
        return self.backend_interface.check_bucket_exists(bucket_name)

    def create_bucket(self, bucket_name, acl=DEFAULT_ACL):
        """Create a new bucket"""
        self.backend_interface.create_bucket(bucket_name, acl)

    def rm_bucket(self, bucket_name):
        '''Remove an empty bucket. Throws an exception when bucket is not empty.
        '''
        self.set_bucket(bucket_name)
        bucket = self.get_bucket()
        try:
            bucket.delete()
        except botocore.exceptions.ClientError as e:
            print("Bucket not empty. To delete, first empty the bucket.")

    def set_bucket(self, bucket_name):
        """Bucket to use"""
        self.backend_interface.set_current_bucket(bucket_name)

    def get_bucket(self):
        """Get bucket boto3 object"""
        return self.backend_interface.get_bucket()

    def get_bucket_objects(self, **kwargs):
        """Get list of objects from the bucket.

        This is a wrapper to ``self.get_bucket().bucket.objects``

        Parameters
        ----------
        limit : int, 1000
            Maximum number of items to return
        page_size : int, 1000
            The page size for pagination
        filter : dict
            A dictionary with key 'Prefix', specifying a prefix
            string.  Only return objects matching this string.
            Defaults to '/' (i.e. all objects).
        kwargs : optional
            Dictionary of {method:value} for ``bucket.objects``

        Returns
        -------
        objects_list : list (boto3 objects)

        Notes
        -----
        If you get a 'PaginationError', this means you have
        a lot of items on your bucket and should increase ``page_size``
        """
        warn('Deprecated. Use get_objects() instead', DeprecationWarning)
        return self.backend_interface.list_objects(**kwargs)

    def get_objects(self, **kwargs):
        """
        Like get_bucket_objects, but more aptly named to the generic interface
        Parameters
        ----------
        self
        kwargs

        Returns
        -------

        """
        return self.backend_interface.list_objects(**kwargs)

    def get_bucket_size(self, limit=10**6, page_size=10**6):
        """Counts the size of all objects in the current bucket.

        Parameters
        ----------
        limit : int, 10^6
            Maximum number of items to return
        page_size : int, 10^6
            The page size for pagination

        Returns
        -------
        total_bytes : int
            The byte count of all objects in the bucket.

        Notes
        -----
        Because paging does not work properly, if there are more than
        limit,page_size number of objects in the bucket, this function will
        underestimate the total size. Check the printed number of objects for
        suspicious round numbers.
        TODO(anunez): Remove this note when the bug is fixed.
        """
        warn('Deprecated, use get_size() instead', DeprecationWarning)
        return self.backend_interface.size

    def get_size(self):
        """
        Gets the total size of the current container of objects. Generic naming.
        Parameters
        ----------
        self

        Returns
        -------

        """
        return self.backend_interface.size

    def show_buckets(self):
        """Show available buckets"""
        self.backend_interface.show_all_buckets()

    @clean_object_name
    def get_object(self, object_name, bucket_name=None):
        """Get a boto3 object. Create it if it doesn't exist"""
        # NOTE: keeping this in case outside code is using this.
        return self.backend_interface.get_s3_object(object_name, bucket_name)

    def show_objects(self, limit=1000, page_size=1000):
        """Print objects in the current bucket"""
        if self.backend == 's3':
            object_list = self.backend_interface.list_objects(limit=limit, page_size=page_size)
            try:
                print_objects(object_list)
            except botocore.exceptions.PaginationError:
                print('Loads of objects in "%s". Increasing page_size by 100x...' % self.bucket_name)
                object_list = self.backend_interface.list_objects(limit = limit, page_size = page_size * 100)
                print_objects(object_list)
        elif self.backend == 'gdrive':
            drivefiles = self.backend_interface.drive.ListFile({'q': "trashed=false"}).GetList()
            object_list = [df['title'] for df in drivefiles]
            for obj in object_list:
                # TODO: also print last modified date and whatever else to match s3
                print(obj)
        else:
            print(self.backend_interface.list_objects())

    # With the abstraction of the cloud interface, and the local interface, all file I/O methods
    # should be calling upload_object() and download_stream() instead of directly interfacing with the
    # CCBackEnd object or the actual cloud APIs

    @clean_object_name
    def upload_object(self, object_name, body, acl=DEFAULT_ACL, **metadata):
        # First check size of object to see if MPU is necessary
        if get_fileobject_size(body) > MPU_THRESHOLD:
            self.mpu_fileobject(object_name, body, acl=acl, **metadata)
        else:
            self.backend_interface.upload_stream(body, object_name, metadata, permissions=acl)

    def download_stream(self, object_name):
        """
        Returns the CloudStream object for an object
        Parameters
        ----------
        self
        object_name

        Returns
        -------
        CloudStream object
        """
        return self.backend_interface.download_stream(object_name)

    def upload_from_file(self, flname, object_name=None,
                         ExtraArgs=dict(ACL=DEFAULT_ACL)):
        """Upload a file to the cloud.

        Parameters
        ----------
        file_name : str
            Absolute path of file to upload
        object_name : str, None
            Name of uploaded object. If None, use
            the full file name as the object name.
        ExtraArgs : dict
            Defaults ``dict(ACL=DEFAULT_ACL)``

        Returns
        -------
        response : boto3 response
        """
        return self.backend_interface.upload_file(flname, object_name, ExtraArgs['ACL'])

    def upload_from_directory(self, disk_path, cloud_path=None,
                              recursive=False, ExtraArgs=dict(ACL=DEFAULT_ACL)):
        '''Upload a directory to the cloud
        '''
        from glob import glob

        filenames = sorted(os.listdir(disk_path))
        if cloud_path is None:
            cloud_path = disk_path
        for flname in filenames:
            flpath = os.path.join(disk_path, flname)
            obname = self.pathjoin(cloud_path, flname)
            if os.path.isfile(flpath):
                self.upload_from_file(flpath, obname, ExtraArgs=ExtraArgs)
            elif os.path.isdir(flpath):
                if recursive:
                    self.upload_from_directory(flpath, obname,
                                               recursive=recursive,
                                               ExtraArgs=ExtraArgs)
        print('Uploaded "%s" to "%s"'%(disk_path, cloud_path))

    @clean_object_name
    def download_to_file(self, object_name, file_name):
        """Download cloud object to a file

        Parameters
        ----------
        object_name : str
        file_name : str
            Absolute path where the data will be downloaded on disk
        """
        return self.backend_interface.download_to_file(object_name, file_name)

    @clean_object_name
    def download_object(self, object_name):
        """Download object raw data.
        This simply calls the object body ``read()`` method.

        Parameters
        ---------
        object_name : str

        Returns
        -------
        byte_data : str
            Object byte contents
        """
        return self.download_stream(object_name).content.read()

    @clean_object_name
    def mpu_fileobject(self, object_name, file_object,
                       buffersize=MPU_CHUNKSIZE, verbose=True, acl=DEFAULT_ACL, **metadata):
        """Multi-part upload for a file-object.

        This automatically creates a multipart upload of an object.
        Useful for large objects that are loaded in memory. This avoids
        having to write the file to disk and then using ``upload_from_file``.

        Parameters
        ----------
        object_name : str
        file_object :
            file-like object (e.g. StringIO, file, etc)
        buffersize  : int, (defaults to 100MB)
            Byte size of the individual parts to create.
        verbose     : bool
            verbosity flag of whether to print mpu information to stdout
        **metadata  : optional
            Metadata to store along with MPU object
        """
        return self.backend_interface.upload_multipart(file_object, object_name, metadata,
                                                       buffersize=buffersize,
                                                       permissions=acl,
                                                       verbose=verbose)



    @clean_object_name
    def upload_json(self, object_name, ddict, acl=DEFAULT_ACL, **metadata):
        """Upload a dict as a JSON using ``json.dumps``

        Parameters
        ----------
        object_name : str
        ddict : dict to upload
        metadata : dict, optional
        """
        json_data = json.dumps(ddict)
        return self.upload_object(object_name, StringIO(json_data.encode()), acl, **metadata)

    @clean_object_name
    def download_json(self, object_name):
        """Download a JSON object

        Parameters
        ----------
        object_name : str

        Returns
        -------
        json_data : dict
            Dictionary representation of JSON file
        """
        self.exists_object(object_name, raise_err=True)
        obj = self.download_object(object_name)
        return json.loads(obj.decode())

    @clean_object_name
    def upload_pickle(self, object_name, data_object, acl=DEFAULT_ACL, **metadata):
        """Upload an object using pickle: ``pickle.dumps``

        Parameters
        ----------
        object_name : str
        data_object : object
        """
        object_to_upload = StringIO(pickle.dumps(data_object))
        response = self.upload_object(object_name, object_to_upload, acl=acl, **metadata)
        return response

    @clean_object_name
    def download_pickle(self, object_name):
        """Download a pickle object

        Parameters
        ----------
        object_name : str

        Returns
        -------
        data_object : object
        """
        self.exists_object(object_name, raise_err=True)
        obj = self.download_object(object_name)
        return pickle.loads(obj)

class ArrayInterface(BasicInterface):
    """Provides numpy.array concepts.
    """

    def __init__(self, *args, **kwargs):
        """
        Parameters
        ----------
        bucket_name : str
        ACCESS_KEY : str
        SECRET_KEY : str
        endpoint_url : str
            The URL for the S3 gateway
        force_bucket_creation : bool
            Create requested bucket if it doesn't exist

        Returns
        -------
        cci : ccio
            Cottoncandy interface object
        """
        super(ArrayInterface, self).__init__(*args, **kwargs)

    @clean_object_name
    def upload_npy_array(self, object_name, array, acl=DEFAULT_ACL, **metadata):
        """Upload a np.ndarray using ``np.save``

        This method creates a copy of the array in memory
        before uploading since it relies on ``np.save`` to
        get a byte representation of the array.

        Parameters
        ----------
        object_name : str
        array : numpy.ndarray
        acl : ACL for this object
        **metadata : extra kwargs are uploaded to object metadata

        Returns
        -------
        reponse : boto3 upload response

        See Also
        --------
        :func:`upload_raw_array` which is more efficient
        """
        # TODO: check array.dtype.hasobject
        arr_strio = StringIO()
        np.save(arr_strio, array)
        arr_strio.seek(0)
        response = self.upload_object(object_name, arr_strio, acl, **metadata)
        return response

    @clean_object_name
    def download_npy_array(self, object_name):
        """Download a np.ndarray uploaded using ``np.save`` with ``np.load``.

        Parameters
        ----------
        object_name : str

        Returns
        -------
        array : np.ndarray
        """
        self.exists_object(object_name, raise_err=True)
        array = np.load(StringIO(self.download_object(object_name)))
        return array

    @clean_object_name
    def upload_raw_array(self, object_name, array, compression=DO_COMPRESSION, acl=DEFAULT_ACL, **metadata):
        """Upload a binary representation of a np.ndarray

        This method reads the array content from memory to upload.
        It does not have any overhead.

        Parameters
        ----------
        object_name : str
        array : np.ndarray
        compression  : str, bool
            `True` uses the configuration defaults. `False` is no compression.
            Available options are: 'gzip', 'LZ4', 'Zlib', 'Zstd', 'BZ2' (attend to caps).
            NB: Zstd appears to be the only one that supports >2GB arrays.
        acl : str
            ACL for the object
        **metadata : optional

        Notes
        -----
        This method also uploads the array ``dtype``, ``shape``, and ``gzip``
        flag as metadata
        """
        if compression is None:
            compression = False

        # Backward compatibility
        if 'gzip' in metadata:
            warn("Deprecated keyword argument `gzip`. Use `compression='gzip'` instead", DeprecationWarning)
            gz = metadata.pop('gzip')
            compression = 'gzip' if gz else False

        if compression is True:
            # check whether array is >= 2 GB
            large_array = array.nbytes > 2**31
            compression = COMPRESSION_LARGE if large_array else COMPRESSION_SMALL

            if large_array and compression == 'gzip':
                # Raise exception for specification of gzip w/ large array
                raise ValueError(("gzip does not support compression of >2GB arrays. "
                                  "Try `compression='Zstd'` instead."))

        order = 'C' if array.flags.carray else 'F'
        if ((not array.flags['%s_CONTIGUOUS' % order] and six.PY2) or
            (not array.flags['C_CONTIGUOUS'] and six.PY3)):
            warn('Non-contiguous array. Creating copy (will use extra memory)...')

            if six.PY3 and order == 'F':
                # memoryview (PY3) vs buffer (PY2) issues
                warn("PY3: Changing array from 'F' to 'C' order")
                order = 'C'

            # create contiguous copy
            array = np.array(array, order=order)

        meta = dict(dtype=array.dtype.str,
                    shape=','.join(map(str, array.shape)),
                    compression=str(compression),
                    order=order)

        # check for conflicts in metadata
        metadata_keys = []
        for k in metadata.keys():
            # check for conflicts in metadata
            metadata_keys.append(k in meta)

        assert not any(metadata_keys)
        meta.update(metadata)

        if compression is False:
            filestream = StringIO(array.data)
        elif compression == 'gzip':
            if six.PY3 and array.flags['F_CONTIGUOUS']:
                # eventually, array.data below should be changed to np.getbuffer(array)
                # (not yet working in python3 numpy)
                # F-contiguous arrays break gzip in python 3
                array = array.T
            zipdata = StringIO()
            gz = GzipFile(mode='wb', fileobj=zipdata)
            gz.write(array.data)
            gz.close()
            zipdata.seek(0)
            filestream = zipdata
        elif hasattr(numcodecs, compression.lower()):
            # If the specified compression type is in numcodecs, use numcodecs
            orig_nbytes = array.nbytes
            compressor = numcodecs.get_codec(dict(id=compression.lower()))
            filestream = StringIO(compressor.encode(array))
            data_nbytes = get_fileobject_size(filestream)
            print('Compressed to %0.2f%% the size'%(data_nbytes / float(orig_nbytes) * 100))
        else:
            raise ValueError('Unknown compression scheme: %s'%compression)
        response = self.upload_object(object_name, filestream, acl=acl, **meta)
        return response

    @clean_object_name
    def download_raw_array(self, object_name, buffersize=2**16, **kwargs):
        """Download a binary np.ndarray and return an np.ndarray object
        This method downloads an array without any disk or memory overhead.

        Parameters
        ----------
        object_name : str
        buffersize  : optional (defaults 2^16)

        Returns
        -------
        array : np.ndarray

        Notes
        -----
        The object must have metadata containing: shape, dtype and a gzip
        boolean flag. This is all automatically handled by ``upload_raw_array``.
        """
        self.exists_object(object_name, raise_err=True)

        arraystream = self.download_stream(object_name)

        shape = arraystream.metadata['shape']
        shape = tuple(map(int, shape.split(',')) if shape else ())
        dtype = np.dtype(arraystream.metadata['dtype'])
        order = arraystream.metadata.get('order', 'C')
        array = np.empty(tuple(shape), dtype = dtype, order = order)

        body = arraystream.content


        if 'gzip' in arraystream.metadata:
            # Backward compatibility, over-write "compression" value
            isgzipped = string2bool(arraystream.metadata['gzip'])
            if isgzipped:
                arraystream.metadata['compression'] = 'gzip'
            else:
                arraystream.metadata['compression'] = 'False'

        assert 'compression' in arraystream.metadata

        if arraystream.metadata['compression'] == 'gzip':
            # gzipped!
            datastream = GzipInputStream(body)
        elif arraystream.metadata['compression'] in ['False', 'None']:
            # uncompressed data
            datastream = body
        else:
            # numcodecs compression
            compression = arraystream.metadata['compression']
            decompressor = numcodecs.get_codec(dict(id=compression.lower()))
            # Can't decode stream; must read in file. Memory hungry?
            bits_compr = arraystream.content.read()
            bits = decompressor.decode(bits_compr)
            array = np.frombuffer(bits, dtype=dtype)
            array = array.reshape(shape)
            return array

        read_buffered(datastream, array, buffersize=buffersize)
        return array

    @clean_object_name
    def dict2cloud(self, object_name, array_dict, acl=DEFAULT_ACL,
                   verbose=True, **metadata):
        """Upload an arbitrary depth dictionary containing arrays

        Parameters
        ----------
        object_name : str
        array_dict  : dict
            An arbitrary depth dictionary of arrays. This can be
            conceptualized as implementing an HDF-like group
        verbose : bool
            Whether to print object_name after completion
        """
        for k, v in array_dict.items():
            name = self.pathjoin(object_name, k)

            if isinstance(v, dict):
                _ = self.dict2cloud(name, v, acl=acl, **metadata)
            elif isinstance(v, np.ndarray):
                _ = self.upload_raw_array(name, v, acl=acl, **metadata)
            else:  # try converting to array
                _ = self.upload_raw_array(name, np.asarray(v), acl=acl)

        if verbose:
            print('uploaded arrays in "%s"' % object_name)

    @clean_object_name
    def cloud2dict(self, object_root, verbose=True, keys=None, **metadata):
        """Download all the arrays of the object branch and return a dictionary.
        This is the complement to ``dict2cloud``

        Parameters
        ----------
        object_root : str
            The branch to create the dictionary from
        verbose : bool
            Whether to print object_root after completion
        keys : A list of strings
            Specify which keys to download

        Returns
        -------
        datadict  : dict
            An arbitrary depth dictionary.
        """
        # TODO: gdrive compatibility?
        datadict = {}

        if keys is not None:
            if isinstance(keys, str):
                keys = [keys]
            subdirs = keys
        else:
            subdirs = self.lsdir(object_root)
            subdirs = [os.path.split(t)[-1] for t in subdirs]


        if not subdirs:
            print('Nothing found in "%s"' % object_root)
            return

        for subdir in subdirs:
            path = self.pathjoin(object_root, subdir)

            if self.exists_object(path):
                # TODO: allow non-array things
                try:
                    arr = self.download_raw_array(path)
                except KeyError as e:
                    print('Could not download "%s: missing %s from metadata"' % (path, e))
                    arr = None
                datadict[subdir] = arr
            else:
                datadict[subdir] = self.cloud2dict(path)

            

        if verbose:
            print('Downloaded arrays in "%s"' % object_root)

        return datadict

    @clean_object_name
    def cloud2dataset(self, object_root, **metadata):
        """Get a dataset representation of the object branch.

        Parameters
        ----------
        object_root : str
            The branch to create a dataset from

        Returns
        -------
        cc_dataset_object  : cottoncandy.BrowserObject
            This can be conceptualized as implementing an h5py/pytables
            object with ``load()`` and ``keys()`` methods.
        """
        from cottoncandy.browser import S3Directory
        return S3Directory(object_root, interface = self)

    @clean_object_name
    def upload_dask_array(self, object_name, arr, axis=-1, buffersize=DASK_CHUNKSIZE, **metakwargs):
        """Upload an array in chunks and store the metadata to reconstruct
        the complete matrix with ``dask``.

        Parameters
        ----------
        object_name : str
        arr  : np.ndarray
        axis : int or None (default: -1)
            The axis along which to slice the array. If None is given,
            the array is chunked into ideal isotropic voxels.
            ``axis=None`` is WIP and atm works fine for near isotropic matrices
        buffersize : scalar (default: 100MB)
            Byte size of the desired array chunks

        Returns
        -------
        response : boto3 response

        Notes
        -----
        Each array chunk is uploaded as a raw np.array with the prefix "pt%04i".
        The metadata is stored as a json file ``metadata.json``. For example, if
        an array is uploaded with the name "my_array_name" and split into 2 parts,
        the following objects are created:

        * my_array_name/pt0000
        * my_array_name/pt0001
        * my_array_name/metadata.json
        """
        metadata = dict(shape = arr.shape,
                        dtype = arr.dtype.str,
                        dask = [],
                        chunk_sizes = [],
                        )

        generator = generate_ndarray_chunks(arr, axis = axis, buffersize = buffersize)
        total_upload = 0.0
        for idx, (chunk_coord, chunk_arr) in enumerate(generator):
            chunk_arr = chunk_arr.copy()
            total_upload += chunk_arr.nbytes
            txt = (idx + 1, total_upload / MB, arr.nbytes / float(MB))
            print('uploading %i: %0.02fMB/%0.02fMB' % txt)

            part_name = self.pathjoin(object_name, 'pt%04i' % idx)
            metadata['dask'].append((chunk_coord, part_name))
            metadata['chunk_sizes'].append(chunk_arr.shape)
            self.upload_raw_array(part_name, chunk_arr)

        # convert to dask convention (sorry)
        details = [t[0] for t in metadata['dask']]
        dimension_sizes = [dict() for idx in range(arr.ndim)]
        for dim, chunks in enumerate(zip(*details)):
            for sample_idx, chunk_idx in enumerate(chunks):
                if chunk_idx not in dimension_sizes[dim]:
                    dimension_sizes[dim][chunk_idx] = metadata['chunk_sizes'][sample_idx][dim]

        chunks = [[value for k, value in sorted(sizes.items())] for sizes in dimension_sizes]
        metadata['chunks'] = chunks
        return self.upload_json(self.pathjoin(object_name, 'metadata.json'), metadata, **metakwargs)

    @clean_object_name
    def download_dask_array(self, object_name, dask_name='array'):
        """Downloads a split matrix as a ``dask.array.Array`` object

        This uses the stored object metadata to reconstruct the full
        n-dimensional array uploaded using ``upload_dask_array``.

        Examples
        --------
        >>> s3_response = cci.upload_dask_array('test_dim', arr, axis=-1)
        >>> dask_object = cci.download_dask_array('test_dim')
        >>> dask_object
        dask.array<array, shape=(100, 600, 1000), dtype=float64, chunksize=(100, 600, 100)>
        >>> dask_slice = dask_object[..., :200]
        >>> dask_slice
        dask.array<getitem..., shape=(100, 600, 1000), dtype=float64, chunksize=(100, 600, 100)>
        >>> downloaded_data = np.asarray(dask_slice) # this downloads the array
        >>> downloaded_data.shape
        (100, 600, 200)
        """
        from dask import array as da

        metadata = self.download_json(self.pathjoin(object_name, 'metadata.json'))
        chunks = tuple(tuple(chunk) for chunk in metadata['chunks'])
        shape = tuple(metadata['shape'])
        dtype = np.dtype(metadata['dtype'])

        dask = {(dask_name,) + tuple(shape): (self.download_raw_array, part_name) \
                for shape, part_name in metadata['dask']}

        return da.Array(dask, dask_name, chunks, shape = shape, dtype = dtype)

    @clean_object_name
    def upload_sparse_array(self, object_name, arr):
        """Uploads a scipy.sparse array as a folder of array objects

        Parameters
        ----------
        object_name : str
            The name of the object to be stored.
        arr : scipy.sparse.spmatrix
            A scipy.sparse array to be saved. If type is DOK or LIL,
            it will be converted to csr before saving
        """
        if isinstance(arr, csr_matrix):
            attrs = ['data', 'indices', 'indptr']
            arrtype = 'csr'
        elif isinstance(arr, coo_matrix):
            attrs = ['row', 'col', 'data']
            arrtype = 'coo'
        elif isinstance(arr, csc_matrix):
            attrs = ['data', 'indices', 'indptr']
            arrtype = 'csc'
        elif isinstance(arr, bsr_matrix):
            attrs = ['data', 'indices', 'indptr']
            arrtype = 'bsr'
        elif isinstance(arr, dia_matrix):
            attrs = ['data', 'offsets']
            arrtype = 'dia'
        else:  # dok and lil: convert to csr and save
            # TODO: warn user here that matrix type will be changed?
            arr = arr.tocsr()
            attrs = ['data', 'indices', 'indptr']
            arrtype = 'csr'

        # Upload parts
        for attr in attrs:
            self.upload_raw_array(self.pathjoin(object_name, attr), getattr(arr, attr))

        # Upload metadata
        metadata = dict(type = arrtype, attrs = attrs, shape = arr.shape)
        return self.upload_json(self.pathjoin(object_name, 'metadata.json'), metadata)

    @clean_object_name
    def download_sparse_array(self, object_name):
        """Downloads a scipy.sparse array

        Parameters
        ----------
        object_name : str
            The object name for the sparse array to be retrieved.

        Returns
        -------
        arr : scipy.sparse.spmatrix
            The array stored at the location given by object_name
        """
        # Get metadata
        metadata = self.download_json(self.pathjoin(object_name, 'metadata.json'))
        # Get type, shape
        arrtype = metadata['type']
        shape = metadata['shape']
        # Get data
        d = dict()
        for attr in metadata['attrs']:
            d[attr] = self.download_raw_array(self.pathjoin(object_name, attr))

        if arrtype == 'csr':
            arr = csr_matrix((d['data'], d['indices'], d['indptr']),
                             shape = shape)
        elif arrtype == 'coo':
            arr = coo_matrix((d['data'], (d['row'], d['col'])),
                             shape = shape)
        elif arrtype == 'csc':
            arr = csc_matrix((d['data'], d['indices'], d['indptr']),
                             shape = shape)
        elif arrtype == 'bsr':
            arr = bsr_matrix((d['data'], d['indices'], d['indptr']),
                             shape = shape)
        elif arrtype == 'dia':
            arr = dia_matrix((d['data'], d['offsets']), shape = shape)

        return arr


class FileSystemInterface(BasicInterface):
    """Emulate some file system functionality.
    """

    def __init__(self, *args, **kwargs):
        """
        Parameters
        ----------
        bucket_name : str
        ACCESS_KEY : str
        SECRET_KEY : str
        endpoint_url : str
            The URL for the S3 gateway
        force_bucket_creation : bool
            Create requested bucket if it doesn't exist

        Returns
        -------
        cci : ccio
            Cottoncandy interface object
        """
        super(FileSystemInterface, self).__init__(*args, **kwargs)

    def lsdir(self, path='/', limit=10**3):
        """List the contents of a directory

        Parameters
        ----------
        path : str (default: "/")

        Returns
        -------
        matches : list
            The children of the path.
        """
        return self.backend_interface.list_directory(path, limit)

    @clean_object_name
    def ls(self, pattern, page_size=10**3, limit=10**3, verbose=False):
        """File-system like search for S3 objects

        Parameters
        ----------
        pattern : str
            A ls-style command line like query

        page_size : int (default: 1,000)
        limit : int (default: 1,000)

        Returns
        -------
        object_names : list
            Object names that match the search pattern

        Notes
        -----
        Increase ``page_size`` and ``limit`` if you have a lot of objects
        otherwise, the search might not return all matching objects in store.
        """
        pattern = remove_trivial_magic(pattern)
        pattern = os.path.normpath(pattern)

        if has_real_magic(pattern):
            magic_check = re.compile('[*?[]')
            prefix = magic_check.split(pattern)[0]
        else:
            prefix = pattern

        # get objects that match common prefix
        if not has_real_magic(pattern):
            object_names = self.lsdir(prefix, limit = limit)
        else:
            object_list = self.get_objects(filter = dict(Prefix = prefix),
                                                  page_size = page_size,
                                                  limit = limit)
            object_names = objects2names(object_list)

        # remove trailing '/'
        object_names = map(os.path.normpath, object_names)

        if has_real_magic(pattern):
            # get the unique sub-directories
            depth = len(pattern.split(SEPARATOR))
            object_names = {SEPARATOR.join(t.split(SEPARATOR)[:depth]): 1 for t in object_names}.keys()
            # filter the list with glob pattern
            object_names = fnmatch.filter(object_names, pattern)
        if verbose:
            print('\n'.join(sorted(object_names)))
        return list(object_names)

    @clean_object_name
    def glob(self, pattern, **kwargs):
        """Return a list of object names in the cloud storage
        that match the glob pattern.

        Parameters
        ----------
        pattern : str,
            A glob pattern string
        verbose : bool, optional
            If True, also print object name and creation date
        limit : None, int, optional
        page_size: int, optional

        Returns
        -------
        object_names : list

        Example
        -------
        >>> cci.glob('/path/to/*/file01*.grp/image_data')
        ['/path/to/my/file01a.grp/image_data',
         '/path/to/my/file01b.grp/image_data',
         '/path/to/your/file01a.grp/image_data',
         '/path/to/your/file01b.grp/image_data']
        >>> cci.glob('/path/to/my/file02*.grp/*')
        ['/path/to/my/file02a.grp/image_data',
         '/path/to/my/file02a.grp/text_data',
         '/path/to/my/file02b.grp/image_data',
         '/path/to/my/file02b.grp/text_data',]

        Extended Summary
        ----------------
        Some gotchas

        limit: None, int, optional
            The maximum number of objects to return
        page_size: int, optional
            This is important for buckets with loads of objects.
            By default, ``glob`` will download a maximum of 10^6
            object names and perform the search. If more objects exist,
            the search might not find them and the page_size should
            be increased.

        Notes
        -----
        If more than 10^6 objects, provide ``page_size=10**7`` kwarg.
        """
        # determine if we're globbing

        if self.backend == 's3':
            return self.glob_s3(pattern, **kwargs)
        elif self.backend == 'gdrive':
            return self.glob_google_drive(pattern)
        else:
            return self.glob_s3(pattern, **kwargs)

    def glob_google_drive(self, pattern):
        """Globbing on google drive

        Parameters
        ----------
        pattern

        Returns
        -------

        """
        matches = []
        if '/' not in pattern:	# end tree, list all objects
            return self.backend_interface.list_objects(True)

        nextFolderExp = r'^/?[^/]*/'
        nextFolder = re.match(nextFolderExp, pattern).group(0)
        pattern = re.sub(nextFolderExp, '', pattern)

        if '*' not in nextFolder:	# no wildcard, simply cd into it
            self.backend_interface.cd(nextFolder)
        else:						# find all wildcard matches and recursively glob them
            files = self.backend_interface.list_objects()
            for f in files:
                matches.append(self.glob_google_drive())

        # TODO: finish this

        return


    def glob_s3(self, pattern, **kwargs):
        """Globbing on S3

        Parameters
        ----------
        pattern
        kwargs

        Returns
        -------

        """
        prefix = MAGIC_CHECK.split(pattern)[0] if has_magic(pattern) else pattern

        page_size = kwargs.get('page_size', 1000000)
        limit = kwargs.get('limit', None)
        do_unquote = kwargs.get('do_unquote', True)

        object_list = self.get_objects(filter = dict(Prefix = prefix),
                                       page_size = page_size,
                                       limit = limit)

        if self.backend == 's3':
            mapper = {obj.key: obj for obj in object_list}
        else:
            mapper = {obj: obj for obj in object_list}
        if do_unquote:
            mapper = {unquote(key): val for key, val in mapper.items()}
        object_names = mapper.keys()

        matches = fnmatch.filter(object_names, pattern) \
            if has_magic(pattern) else object_names
        matches = sorted(matches)

        if kwargs.get('verbose', False):
            # print objects found
            objects_found = [mapper[match] for match in matches]
            if len(objects_found):
                print_objects(objects_found)
            print('Found %i objects matching "%s"' % (len(objects_found), pattern))
        return matches

    @clean_object_name
    def download_directory(self, directory, disk_name):
        """
        Download an entire directory
        NOTE: currently only tested on s3

        Parameters
        ----------
        self
        directory : str
            directory on s3 to download
        disk_name :
            name of directory on disk to download to

        Returns
        -------

        """
        if has_real_magic(directory):
            raise NotImplementedError('Wildcards not implemented')

        if (directory != '') and (directory != '/'):
            directory = remove_root(directory)
        directory = remove_trivial_magic(directory)
        directory = mk_aws_path(directory)

        if not os.path.exists(disk_name):
            os.mkdir(disk_name)

        files = self.glob(directory + '*')
        for f in files:
            if f[-1] == '/':
                continue
            subpath = re.sub(directory, '', f)
            path = os.path.join(disk_name, subpath)
            subfolder = re.match('.*\/', path).group(0)
            if not os.path.exists(subfolder):
                os.makedirs(subfolder)
            self.download_to_file(f, path)

    @clean_object_name
    def search(self, pattern, **kwargs):
        """Print the objects matching the glob pattern

        See ``glob`` documentation for details
        """
        matches = self.glob(pattern, verbose=True, **kwargs)

    def get_browser(self):
        """Return an object which can be tab-completed
        to browse the contents of the bucket as if it were a file-system

        See documentation for ``cottoncandy.get_browser``
        """
        return cottoncandy.browser.S3Directory('', interface = self)

    def cp(self, source_name, dest_name,
           source_bucket=None, dest_bucket=None, overwrite=False):
        """Copy an object

        Parameters
        ----------
        source_name : str
            Name of object to be copied
        dest_name : str
            Copy name
        source_bucket : str
            If copying from a bucket different from the default.
            Defaults to ``self.bucket_name``
        dest_bucket : str
            If copying to a bucket different from the source bucket.
            Defaults to ``source_bucket``
        overwrite : bool (defaults to False)
            Whether to overwrite the `dest_name` object if it already exists
        """
        # TODO: support directories
        return self.backend_interface.copy(source_name, dest_name, source_bucket, dest_bucket, overwrite)

    def mv(self, source_name, dest_name,
           source_bucket=None, dest_bucket=None, overwrite=False):
        """Move an object (make copy and delete old object)

        Parameters
        ----------
        source_name : str
            Name of object to be moved
        dest_name : str
            New object name
        source_bucket : str
            If moving object from a bucket different from the default.
            Defaults to ``self.bucket_name``
        dest_bucket : str (defaults to None)
            If moving to another bucket, provide the bucket name.
            Defaults to ``source_bucket``
        overwrite : bool (defaults to False)
            Whether to overwrite the `dest_name` object if it already exists.
        """
        # TODO: Support directories
        return self.backend_interface.move(source_name, dest_name, source_bucket, dest_bucket, overwrite)

    def rm(self, object_name, recursive=False, delete=True):
        """Delete an object, or a subtree ('path/to/stuff').

        Parameters
        ----------
        object_name : str
            The name of the object to delete. It can also
            be a subtree
        recursive : bool
            When deleting a subtree, set ``recursive=True``. This is
            similar in behavior to 'rm -r /path/to/directory'.
        delete : bool
            When in google drive, actually delete the file or only trash it?

        Example
        -------
        >>> import cottoncandy as cc
        >>> cci = cc.get_interface('mybucket', verbose=False)
        >>> response = cci.rm('data/experiment/file01.txt')
        >>> cci.rm('data/experiment')
        cannot remove 'data/experiment': use `recursive` to remove branch
        >>> cci.rm('data/experiment', recursive=True)
        deleting 15 objects...
        """

        # not moving this to the basic S3Client because it depends on glob
        if self.backend == "s3":
            if self.exists_object(object_name):
                return self.backend_interface.get_s3_object(object_name).delete()
        elif self.backend == "gdrive":
            return self.backend_interface.delete(object_name, recursive, delete)
        else:
            return self.backend_interface.delete(object_name, recursive, delete)

        has_objects = len(self.ls(object_name)) > 0
        if has_objects:
            if recursive:
                all_objects = self.glob(object_name)
                print('deleting %i objects...' % len(all_objects))
                for obname in self.glob(object_name):
                    _ = self.get_object(obname).delete()
                return

        msg = "cannot remove '%s': use `recursive` to remove branch" \
            if has_objects else \
            "nothing found under '%s"
        print(msg % object_name)

    def get_object_owner(self, object_name):
        self.exists_object(object_name, raise_err=True)
        ob = self.get_object(object_name)
        try:
            acl = ob.Acl()
            info = acl.owner
        except botocore.exceptions.ClientError as e:
            if 's3cmd-attrs' in ob.metadata:
                info = ob.metadata['s3cmd-attrs'].split('/')
                info = dict(map(lambda x: x.split(':'), info))
            else:
                raise e
        print(info)


class DefaultInterface(FileSystemInterface,
                       ArrayInterface,
                       BasicInterface):
    """Default cottoncandy interface to the cloud

    This includes numpy.array and file-system-like
    concepts for easy data I/O and bucket/object exploration.
    """

    def __init__(self, *args, **kwargs):
        """
        Parameters
        ----------
        bucket_name : str
        ACCESS_KEY : str
        SECRET_KEY : str
        endpoint_url : str
            The URL for the S3 gateway
        force_bucket_creation : bool
            Create requested bucket if it doesn't exist
        backend : 's3'|'gdrive'
            which backend to hook on to

        Returns
        -------
        cci  : cottoncandy.InterfaceObject
        """
        super(DefaultInterface, self).__init__(*args, **kwargs)




'''Helper functions
'''
import os
import re
import six
import zlib
import string
import urllib
import itertools
from dateutil.tz import tzlocal
from functools import wraps


try:
    from urllib import unquote
except ImportError:
    from urllib.parse import unquote


import numpy as np

from cottoncandy import options

##############################
# Globals
##############################



# S3 AWS
#---------
MB = 2**20
MIN_MPU_SIZE = int(options.config.get('upload_settings', 'min_mpu_size'))*MB # 5MB
MAX_PUT_SIZE = int(options.config.get('upload_settings', 'max_put_size'))*MB # 5GB
MAX_MPU_SIZE = int(options.config.get('upload_settings', 'max_mpu_size_TB'))*MB*MB # 5TB
MAX_MPU_PARTS = int(options.config.get('upload_settings', 'max_mpu_parts')) # 10,000
MPU_THRESHOLD = int(options.config.get('upload_settings', 'mpu_use_threshold'))*MB
MPU_CHUNKSIZE = int(options.config.get('upload_settings', 'mpu_chunksize'))*MB
DASK_CHUNKSIZE = int(options.config.get('upload_settings', 'dask_chunksize'))*MB

SEPARATOR = options.config.get('basic', 'path_separator')

DEFAULT_ACL = options.config.get('basic', 'default_acl')
MANDATORY_BUCKET_PREFIX = options.config.get('basic', 'mandatory_bucket_prefix')

ISBOTO_VERBOSE = options.config.get('login', 'verbose_boto')

##############################
# misc functions
##############################

def sanitize_metadata(metadict):
    outdict = {}
    for key,val in metadict.items():
        outdict[key.lower()] = val
    return outdict

def pathjoin(a, *p):
    """Join two or more pathname components, inserting SEPARATOR as needed.
    If any component is an absolute path, all previous path components
    will be discarded.  An empty last part will result in a path that
    ends with a separator."""
    path = a
    for b in p:
        if b.startswith(SEPARATOR):
            path = b
        elif path == '' or path.endswith(SEPARATOR):
            path += b
        else:
            path += SEPARATOR + b
    return path

def string2bool(mstring):
    '''
    '''
    truth_value = False
    if mstring in ['True','true', 'tru', 't',
                   'y','yes', '1']:
        truth_value = True
    elif mstring == 'None':
        truth_value = None
    return truth_value


def bytes2human(nbytes):
    '''Return string representation of bytes.

    Parameters
    ----------
    nbytes : int
        Number of bytes

    Returns
    -------
    human_bytes : str
        Human readable byte size (e.g. "10.00MB", "1.24GB", etc.).
    '''
    if nbytes == 0:
        return '0.00B'
    mapper = {0 : 'B',
              10 : 'KB',
              20 : 'MB',
              30 : 'GB',
              40 : 'TB',
              50 : 'PB',
              60 : 'EB',
              70 : 'ZB',
              }

    exps = sorted(mapper.keys())
    exp_coeff = np.log2(nbytes)
    exp_closest = int((exps[np.abs(exps - exp_coeff).argmin()]))
    if np.log10(nbytes/2.**exp_closest) < 0:
        exp_closest -= 10
    return '%0.02f%s'%(nbytes/2.**exp_closest, mapper[exp_closest])


def get_object_size(boto_s3_object):
    '''Return the size of the S3 object in MB

    Parameters
    ----------
    boto_s3_object : boto object

    Returns
    -------
    object_size : float (in MB)
    '''
    boto_s3_object.load()
    return boto_s3_object.meta.data['ContentLength']/2.**20


def get_fileobject_size(file_object):
    '''Return byte size of file-object

    Parameters
    ----------
    file_object : file object

    Returns
    -------
    nbytes : int
    '''
    file_object.seek(0,2)
    nbytes = file_object.tell()
    file_object.seek(0)
    return nbytes


def get_key_from_s3fs():
    '''Get AWS keys from default S3fs location if available.

    Returns
    -------
    ACCESS_KEY : str
    SECRET_KEY : str

    Notes
    -----
    Reads ~/.passwd-s3fs to get ACCESSKEY and SECRET KEY
    '''
    key_path = os.path.expanduser('~/.passwd-s3fs')
    if os.path.exists(key_path):
        with open(key_path, 'r') as kfl:
            content = kfl.readline()
            ACCESS_KEY, SECRET_KEY = content.strip().split(':')
            return ACCESS_KEY, SECRET_KEY


def get_key_from_environ():
    '''Get AWS keys from environmental variables if available

    Returns
    -------
    ACCESS_KEY : str
    SECRET_KEY : str

    Notes
    -----
    Reads AWS_ACCESS_KEY and AWS_SECRET_KEY
    '''
    try:
        return os.environ['AWS_ACCESS_KEY'], os.environ['AWS_SECRET_KEY']
    except:
        return


def get_keys():
    '''Read AWS keys from S3fs configuration or environmental variables.

    Returns
    -------
    ACCESS_KEY : str
    SECRET_KEY : str
    '''
    # try to outload keys
    resulta = get_key_from_s3fs()
    resultb = get_key_from_environ()
    result = resultb if (resulta is None) else resulta
    assert result is not None
    access_key, secret_key = result
    return access_key, secret_key


##############################
# object display
##############################

def objects2names(objects):
    '''Return the name of all objects in a list

    Parameters
    ----------
    objects : list (of boto3 objects)

    Returns
    -------
    object_names : list (of strings)
    '''
    return [unquote(t.key) for t in objects]


def unquote_names(object_names):
    '''Clean URL names from a list.

    Parameters
    ----------
    object_names : list (of strings)

    Returns
    -------
    clean_object_names : list (of strings)
    '''
    return [unquote(t) for t in object_names]


def print_objects(object_list):
    '''Print name, size, and creation date of objects in list.

    Parameters
    ----------
    object_list : list (of boto3 objects)
    '''
    object_names = objects2names(object_list)
    if len(object_names):
        maxlen = max(map(len, object_names))
        dates = [t.last_modified.astimezone(tzlocal()).strftime('%Y/%m/%d (%H:%M:%S)')\
                 for t in object_list]
        padding = '{0: <%i} {1} {2}M'%(min(maxlen+3, 70))
        sizes = [round(t.meta.data['Size']/2.**20,1) for t in object_list]
        info = [padding.format(name[-100:],date,size) for name,date,size in zip(object_names, dates, sizes)]
        print('\n'.join(info))


# object naming convention
##############################


def clean_object_name(input_function):
    '''Remove leading "/" from object_name

    This is important for compatibility with S3fs.
    S3fs does not list objects with a "/" prefix.
    '''
    @wraps(input_function)
    def iremove_root(self, object_name, *args, **kwargs):
        object_name = re.sub('//+', '/', object_name)

        if object_name == '':
            pass
        elif object_name[0] == SEPARATOR:
            object_name = object_name[1:]

        return input_function(self, object_name, *args, **kwargs)
    return iremove_root


def remove_root(string_):
    '''remove leading "/" from a string'''
    if string_[0] == SEPARATOR:
        string_ = string_[1:]
    return string_


##############################
# path handling
##############################

check_digits = re.compile('[0-9]')

def has_start_digit(s):
    return check_digits.match(s) is not None

MAGIC_CHECK = re.compile('[*?[]')

def has_magic(s):
    '''Check string to see if it has any glob magic
    '''
    return MAGIC_CHECK.search(s) is not None


def has_trivial_magic(s):
    '''Check string to see if it has trivial glob magic
    (e.g. "path/*")
    '''
    trivial_magic = re.compile('[^*]*/[\*]$') # "/*" at end only
    istrivial = trivial_magic.match(s) is not None
    magic_check = re.compile('[?[]')
    ismagic = magic_check.search(s) is not None
    if istrivial and (not ismagic):
        return True
    else:
        return False


def has_real_magic(s):
    '''Check if string has non-trivial glob pattern
    '''
    return has_magic(s) and (not has_trivial_magic(s))


def remove_trivial_magic(s):
    '''
    * xxx/*      -> xxx/
    * xxx/       -> xxx/
    * xxx/*/yyy/ -> xxx/*/yyy/
    '''
    if has_real_magic(s) or (not has_magic(s)):
        return s
    assert has_trivial_magic(s)
    return s[:-1] # remove '*' at end


def split_uri(uri, pattern='s3://', separator='/'):
    """Convert a URI to a bucket, object name tuple.

    's3://bucket/path/to/thing' -> ('bucket', 'path/to/thing')
    """
    assert pattern in uri
    parts = uri[len(pattern):].split(separator)
    bucket = parts[0]
    path = separator.join(parts[1:])
    return bucket, path


def mk_aws_path(path):
    """Make the `path` behave as expected when querying S3 with
    `list_objects`.

    * xxx/yyy -> xxx/yyy/
    * xxx/    -> xxx/
    * xxx     -> xxx/
    * /       -> ''
    * ''      -> ''
    """
    if (path == SEPARATOR) or (path == ''):
        return ''

    matcher = re.compile('/$') # at end only
    if matcher.search(path) is not None:
        return path
    else:
        return path + SEPARATOR


##############################
# low-level array handling
##############################


def generate_ndarray_chunks(arr, axis=None, buffersize=100*MB):
    '''A generator that splits an array into chunks of desired byte size

    Parameters
    ----------
    arr  : np.ndarray
    axis : int, None
        The axis along which to slice the array. If None is given,
        the array is chunked into ideal isotropic voxels.
    buffersize : scalar
        Byte size of the desired array chunks

    Returns
    -------
    iterator : generator object
        The object yields the tuple:
        (chunk_coordinates, chunk_data_slice)

        * chunk_coordinates:
          Indices of the current chunk along each dimension
        * chunk_data_slice:
          Data for this chunk
    Notes
    -----
    ``axis=None`` is WIP and only works well for near isotropic matrices.
    '''
    shape = arr.shape
    nbytes_total = arr.nbytes
    nparts = int(np.ceil(nbytes_total / float(buffersize)))

    # make sure we can upload
    assert nbytes_total < MAX_MPU_SIZE # 5TB
    assert buffersize >= MIN_MPU_SIZE  # 5MB
    assert nparts < MAX_MPU_PARTS      # 10,000

    if axis is None:
        # split array into ideal isotropic chunks
        logsum = 0
        factor = arr.ndim
    else:
        if axis == -1: axis = (arr.ndim - 1)
        dims = [arr.shape[t] for t in range(arr.ndim) if t != axis]
        logsum = np.sum([np.log(t) for t in  dims])
        factor = 1

    logii = ((np.log(buffersize) - np.log(arr.itemsize)) - logsum)/factor
    ii = int(np.ceil(np.exp(logii)))
    dim_nchunks = map(lambda x: int(np.ceil(x/ii)) + 1, shape)

    if axis is not None:
        # only slicing one dimension
        dim_nchunks = [(t if (i == axis) else 1) for i,t in enumerate(dim_nchunks)]

    chunk_shapes = [ii if (axis is None) else \
                    (arr.shape[dim] if (dim != axis) else ii) \
                    for dim in range(arr.ndim)]

    dim_ranges = map(lambda x: range(x), dim_nchunks)
    iterator = itertools.product(*dim_ranges)
    for chunk_idx, chunk_coords in enumerate(iterator):
        beg = [chunk_shapes[dim]*cc for dim, cc in enumerate(chunk_coords)]
        end = [min(chunk_shapes[dim]*(cc+1), shape[dim])for dim, cc in enumerate(chunk_coords)]
        slicers = tuple(map(lambda dim_lims: slice(dim_lims[0],dim_lims[1]), zip(beg,end)))
        yield chunk_coords, arr[slicers]


def read_buffered(frm, to, buffersize=64):
    '''Fill a numpy n-d array with file-like object contents

    Parameters
    ----------
    frm : buffer
        Object with a ``read`` method
    to : np.ndarray
        Array to which the contents will be put
    '''
    nbytes_total = to.size * to.dtype.itemsize
    if six.PY3:
        if to.flags['F_CONTIGUOUS']:
            vw = to.T.view()
        else:
            vw = to.view()
        vw.shape = (-1,)                # Must be a ravel-able object
        vw.dtype = np.dtype('uint8')    # 256 values in each byte

    for ci in range(int(np.ceil(nbytes_total / float(buffersize)))):
        start = ci * buffersize
        end = min(nbytes_total, (ci + 1) * buffersize)
        if six.PY2:
            to.data[start:end] = frm.read(end - start)
        elif six.PY3:
            vw.data[start:end] = frm.read(end - start)
        else:
            raise("Unknown python version") # not sure six will ever do anything here (6=2x3)

class GzipInputStream(object):
    """Simple class that allow streaming reads from GZip files
    (from https://gist.github.com/beaufour/4205533).

    Python 2.x gzip.GZipFile relies on .seek() and .tell(), so it
    doesn't support this (@see: http://bo4.me/YKWSsL).

    Adapted from: http://effbot.org/librarybook/zlib-example-4.py
    """

    def __init__(self, fileobj, block_size=16384):
        """
        Initialize with the given file-like object.

        @param fileobj: file-like object,
        """
        self.BLOCK_SIZE = block_size                       # Read block size
        # zlib window buffer size, set to gzip's format
        self.WINDOW_BUFFER_SIZE = 16 + zlib.MAX_WBITS

        self._file = fileobj
        self._zip = zlib.decompressobj(self.WINDOW_BUFFER_SIZE)
        self._offset = 0  # position in unzipped stream
        self._data = b''

    def __fill(self, num_bytes):
        """
        Fill the internal buffer with 'num_bytes' of data.

        @param num_bytes: int, number of bytes to read in (0 = everything)
        """

        if not self._zip:
            return

        while not num_bytes or len(self._data) < num_bytes:
            data = self._file.read(self.BLOCK_SIZE)
            if not data:
                self._data = self._data + self._zip.flush()
                self._zip = None  # no more data
                break

            self._data = self._data + self._zip.decompress(data)

    def __iter__(self):
        return self

    def seek(self, offset, whence=0):
        if whence == 0:
            position = offset
        elif whence == 1:
            position = self._offset + offset
        else:
            raise IOError("Illegal argument")
        if position < self._offset:
            raise IOError("Cannot seek backwards")

        # skip forward, in blocks
        while position > self._offset:
            if not self.read(min(position - self._offset, self.BLOCK_SIZE)):
                break

    def tell(self):
        return self._offset

    def read(self, size=0):
        self.__fill(size)
        if size:
            data = self._data[:size]
            self._data = self._data[size:]
        else:
            data = self._data
            self._data = b''
        self._offset = self._offset + len(data)
        return data

    def next(self):
        line = self.readline()
        if not line:
            raise StopIteration()
        return line

    def readline(self):
        # make sure we have an entire line
        while self._zip and "\n" not in self._data:
            self.__fill(len(self._data) + 512)

        pos = string.find(self._data, "\n") + 1
        if pos <= 0:
            return self.read()
        return self.read(pos)

    def readlines(self):
        lines = []
        while True:
            line = self.readline()
            if not line:
                break
            lines.append(line)
        return lines



from abc import ABCMeta, abstractmethod

class FileNotFoundError(RuntimeError):
    """File not found error"""

class CCBackEnd(object):
    """
    Interface for cottoncandy backends
    """
    __metaclass__ = ABCMeta
    def __init__(self):
        pass

    ## Basic File IO
    @abstractmethod
    def check_file_exists(self, file_name, bucket_name):
        """Checks whether a file exists on the cloud

        Parameters
        ----------
        file_name : str
            file on cloud
        bucket_name : str
            (s3) bucket to check in

        Returns
        -------
        bool
        """
        pass

    @abstractmethod
    def upload_stream(self, stream, cloud_name, metadata, permissions):
        """Uploads a stream object with a .read() function

        Parameters
        ----------
        stream : stream
            streaming object
        cloud_name : str
            name to use on cloud
        metadata : dict
            custom metadata for this file
        permissions : str?
            permissions for this file

        Returns
        -------
        bool, upload success
        """
        pass

    @abstractmethod
    def upload_file(self, file_name, cloud_name, permissions):
        """Uploads a file from disk

        Parameters
        ----------
        file_name : str
            name of file to upload
        cloud_name : str
            name to use on the cloud
        permissions : str?
            permissions for this file

        Returns
        -------
        bool, upload success
        """
        pass

    @abstractmethod
    def upload_multipart(self, stream, cloud_name, metadata, permissions, buffersize, verbose):
        """Multi-part upload for large stream objects

        Parameters
        ----------
        stream : stream
            streaming object
        cloud_name : str
            name to use on cloud
        metadata : dict
            custom metadata
        permissions : str?
            permissions for this file
        buffersize : int
            s3 uploading buffersize
        verbose : bool
            s3 verbosity

        Returns
        -------
        bool, upload success
        """
        pass

    @abstractmethod
    def download_stream(self, cloud_name):
        """Downloads a object to an in-memory stream

        Parameters
        ----------
        cloud_name : str
            name of object to download

        Returns
        -------
        CloudStream object
        """
        pass

    @abstractmethod
    def download_to_file(self, cloud_name, file_name):
        """Downloads an object directly to disk

        Parameters
        ----------
        cloud_name : str
            name of object to download
        file_name : str
            name on disk to use

        Returns
        -------
        bool, download success
        """
        pass

    ## Basic File management

    @abstractmethod
    def list_directory(self, path, limit):
        """Lists the content of a directory

        Parameters
        ----------
        path : str
            path on the cloud to get contents for

        Returns
        -------

        """
        pass

    @abstractmethod
    def list_objects(self):
        """Gets all objects contained by backend

        Returns
        -------

        """
        pass

    @abstractmethod
    def copy(self, source, destination, source_bucket, destination_bucket, overwrite):
        """Copies an object

        Parameters
        ----------
        source : str
            origin path
        destination : str
            destination path
        source_bucket : str
            (s3) origin bucket
        destination_bucket : str
            (s3) destination bucket
        overwrite : bool
            overwrite if destination exists?

        Returns
        -------
        bool, copy success
        """
        pass

    @abstractmethod
    def move(self, source, destination, source_bucket, destination_bucket, overwrite):
        """Moves an object

        Parameters
        ----------
        source
        destination
        source_bucket
        destination_bucket
        overwrite

        Returns
        -------

        """
        pass

    @abstractmethod
    def delete(self, file_name, recursive=False, delete=False):
        """Deletes an object

        Parameters
        ----------
        file_name : str
            name of cloud object to delete
        recursive : bool
            recursively delete directory?
        delete : bool
            (gdrive) hard delete the file(s)?

        Returns
        -------

        """
        pass

    @property
    @abstractmethod
    def size(self):
        """Size of stored cloud items in bytes

        Returns
        -------
        int
        """
        pass

class CloudStream(object):
    """
    A simple unified representation of an object downloaded from the cloud.
     .content is a streaming object with a .read() function
     .metadata is a dictionary of the custom metadata of this object

    TODO: unified metadata
    """
    def __init__(self, stream, metadata):
        self.content = stream
        self.metadata = metadata



#!/usr/bin/env python
import re
import sys

try:
    import configparser
except ImportError:
    import ConfigParser as configparser


if len(set(('develop', 'bdist_wheel', 'bdist_egg', 'bdist_rpm', 'bdist',
            'sdist', 'bdist_wheel', 'bdist_dumb',
            'bdist_wininst', 'install_egg_info', 'egg_info', 'easy_install')).intersection(sys.argv)) > 0:
    # monkey patch distutils
    from setuptools import setup
    from setuptools.command.install import install
else:
    # use standard library
    from distutils.command.install import install
    from distutils.core import setup

# get version from cottoncandy/__init__.py
__version__ = 0.0
with open('cottoncandy/__init__.py') as f:
    infos = f.readlines()
for line in infos:
    if "__version__" in line:
        match = re.search(r"__version__ = ['\"]([^'\"]*)['\"]", line)
        __version__ = match.groups()[0]


def set_default_options(optfile):
    import os
    import pwd

    config = configparser.ConfigParser()
    config.read(optfile)
    with open(optfile, 'w') as fp:
        config.write(fp)
    print('cottoncandy configuration file: %s'%optfile)


class my_install(install):
    def run(self):
        install.run(self)
        optfile = [f for f in self.get_outputs() if 'defaults.cfg' in f]
        set_default_options(optfile[0])


if 'extra_setuptools_args' not in globals():
    extra_setuptools_args = dict()

long_description = """
A python scientific library for storing and accessing numpy array data on S3. This is achieved by reading arrays from memory and downloading arrays directly into memory. This means that you don't have to download your array to disk, and then load it from disk into your python session."""

def main(**kwargs):
    setup(name="""cottoncandy""",
          version=__version__,
          description="""sugar for S3""",
          author='Anwar O. Nunez-Elizalde',
          author_email='anwarnunez@gmail.com',
          url='http://gallantlab.github.io/cottoncandy/',
          packages=['cottoncandy',
                    ],
          package_data={
              'cottoncandy':[
                  'defaults.cfg',
                ],
              },
          cmdclass=dict(install=my_install),
          include_package_data=True,
          long_description=long_description,
          install_requires=['six', 'botocore', 'boto3', 'python-dateutil',
                            'PyDrive'],
          **kwargs)

if __name__ == "__main__":
    main(**extra_setuptools_args)



# emacs: -*- mode: python; py-indent-offset: 4; indent-tabs-mode: nil -*-
# vi: set ft=python sts=4 ts=4 sw=4 et:
from __future__ import print_function
from functools import reduce
import boto3
import botocore
import logging
from io import BytesIO

from botocore.utils import fix_s3_host
from cottoncandy.utils import *
from .backend import CCBackEnd, CloudStream

try:
    from urllib import unquote
except ImportError:
    from urllib.parse import unquote
try:
    import cPickle as pickle
except ImportError:
    import pickle


class S3Client(CCBackEnd):
    """
    S3 client interface refactored out of cotton candy interfaces to allow for switching between
    S3 and Google Drive
    """

    @staticmethod
    def connect(ACCESS_KEY, SECRET_KEY, url, **kwargs):
        """Connect to S3 using boto

        Parameters
        ----------
        ACCESS_KEY
        SECRET_KEY
        url
        kwargs : dict
            Extra keyword arguments to `boto3.resource`

        Returns
        -------

        """
        s3 = boto3.resource('s3',
                            endpoint_url=url,
                            aws_access_key_id=ACCESS_KEY,
                            aws_secret_access_key=SECRET_KEY,
                            **kwargs)
        s3.meta.client.meta.events.unregister('before-sign.s3', fix_s3_host)
        return s3

    def __init__(self, bucket, access_key, secret_key, s3url, force_bucket_creation=False, **kwargs):
        """Constructor

        Parameters
        ----------
        bucket
        access_key
        secret_key
        s3url
        force_bucket_creation
        """
        super(S3Client, self).__init__()

        self.connection = S3Client.connect(access_key, secret_key, s3url, **kwargs)
        self.url = s3url
        self.bucket_name = None

        if bucket:
            # bucket given
            if self.check_bucket_exists(bucket):
                self.set_current_bucket(bucket)
            elif force_bucket_creation:
                print('* Creating "%s" bucket...\n' % bucket)
                self.create_bucket(bucket)
            else:
                print('* BUCKET "%s" DOES NOT EXIST...\n' % bucket)

        if self.bucket_name is None:
            print('* cottoncandy instantiated without bucket...\n' \
                  '* Use with caution!\n' \
                  '* Many features will not work!!!\n')

        if string2bool(ISBOTO_VERBOSE) is False:
            logging.getLogger('boto3').setLevel(logging.WARNING)
            logging.getLogger('botocore').setLevel(logging.WARNING)

    def get_bucket_name(self, bucket_name):
        """

        Parameters
        ----------
        bucket_name

        Returns
        -------

        """
        bucket_name = self.bucket_name \
            if bucket_name is None \
            else bucket_name
        return bucket_name

    @clean_object_name
    def check_file_exists(self, object_name, bucket_name=None):
        """Check whether object exists in bucket

        Parameters
        ----------
        object_name : str
            The object name
        bucket_name
        """
        bucket_name = self.get_bucket_name(bucket_name)
        ob = self.connection.Object(key = object_name, bucket_name = bucket_name)

        try:
            ob.load()
        except botocore.exceptions.ClientError as e:
            if e.response['Error']['Code'] == "404":
                exists = False
            else:
                raise e
        else:
            exists = True
        return exists

    def check_bucket_exists(self, bucket_name):
        """Check whether the bucket exists

        Parameters
        ----------
        bucket_name

        Returns
        -------

        """
        try:
            self.connection.meta.client.head_bucket(Bucket = bucket_name)
        except botocore.exceptions.ClientError as e:
            if e.response['Error']['Code'] in ["404", "403"]:
                exists = False
            else:
                raise e
        else:
            exists = True
        return exists

    def create_bucket(self, bucket_name, acl=DEFAULT_ACL):
        """Create a new bucket

        Parameters
        ----------
        bucket_name
        acl

        Returns
        -------

        """
        if MANDATORY_BUCKET_PREFIX:
            tt = len(MANDATORY_BUCKET_PREFIX)
            assert bucket_name[:tt] == MANDATORY_BUCKET_PREFIX

        self.connection.create_bucket(Bucket = bucket_name, ACL = acl)
        self.set_current_bucket(bucket_name)

    def set_current_bucket(self, bucket_name):
        """Sets which bucket to use

        Parameters
        ----------
        bucket_name

        Returns
        -------

        """
        if not self.check_bucket_exists(bucket_name):
            raise IOError('Bucket "%s" does not exist' % bucket_name)
        self.bucket_name = bucket_name

    def get_bucket(self):
        """Get bucket boto3 object

        Returns
        -------

        """
        s3_bucket = self.connection.Bucket(self.bucket_name)
        return s3_bucket

    def list_objects(self, **kwargs):
        """Get list of objects from the bucket
        This is a wrapper to ``self.get_bucket().bucket.objects``

        Parameters
        ----------
        limit : int, 1000
            Maximum number of items to return
        page_size : int, 1000
            The page size for pagination
        filter : dict
            A dictionary with key 'Prefix', specifying a prefix
            string.  Only return objects matching this string.
            Defaults to '/', all objects.
        kwargs : optional
            Dictionary of {method:value} for ``bucket.objects``

        Notes
        -----
        If you get a 'PaginationError', this means you have
        a lot of items on your bucket and should increase ``page_size``
        """
        defaults = dict(limit = 1000,
                        page_size = 1000,
                        filter = dict(Prefix = SEPARATOR),
                        )
        defaults.update(kwargs)
        bucket = self.get_bucket()
        prefix = defaults.pop('filter')

        if prefix['Prefix'] == SEPARATOR:
            request = bucket.objects
        else:
            request = bucket.objects.filter(**prefix)

        for method_name, value in defaults.items():
            if value is None:
                continue
            method = getattr(request, method_name)
            if isinstance(value, dict):
                request = method(**value)
            else:
                request = method(value)

        response = request.all()
        return response

    @property
    def size(self):
        return self.get_current_bucket_size()

    def get_current_bucket_size(self, limit=10 ** 6, page_size=10 ** 6):
        """Counts the size of all objects in the current bucket.

        Parameters
        ----------
        limit : int, 1000
            Maximum number of items to return
        page_size : int, 1000
            The page size for pagination

        Returns
        -------
        total_bytes : int
            The byte count of all objects in the bucket.

        Notes
        -----
        Because paging does not work properly, if there are more than
        limit,page_size number of objects in the bucket, this function will
        underestimate the total size. Check the printed number of objects for
        suspicious round numbers.
        TODO(anunez): Remove this note when the bug is fixed.
        """
        assert self.check_bucket_exists(self.bucket_name)
        obs = self.list_objects(limit = limit, page_size = page_size)
        object_sizes = [t.size for t in obs]
        total_bytes = sum(object_sizes)
        num_objects = len(object_sizes)
        del object_sizes

        txt = "%i bytes (%s) over %i objects"
        print(txt % (total_bytes, bytes2human(total_bytes), num_objects))
        return total_bytes

    def show_all_buckets(self):
        """


        Returns
        -------

        """
        all_buckets = list(self.connection.buckets.all())
        timeformat = lambda x: x.creation_date.astimezone(tzlocal()).strftime
        info = ['{0: <40} {1}'.format(unquote(t.name),
                                      timeformat(t)('%Y/%m/%d (%H:%M:%S)')) \
                for t in all_buckets]
        print('\n'.join(info))

    @clean_object_name
    def get_s3_object(self, object_name, bucket_name=None):
        """Get a boto3 object. Create it if it doesn't exist

        Parameters
        ----------
        object_name
        bucket_name

        Returns
        -------

        """
        bucket_name = self.get_bucket_name(bucket_name)
        return self.connection.Object(bucket_name = bucket_name, key = object_name)

    def upload_stream(self, body, object_name, metadata, permissions=DEFAULT_ACL):
        """Uploads a stream

        Parameters
        ----------
        object_name
        body : stream
        permissions
        metadata

        Returns
        -------

        """
        obj = self.get_s3_object(object_name)
        return obj.put(Body=body, ACL=permissions, Metadata=metadata)

    def download_stream(self, object_name):
        """Download object raw data.
        This simply calls the object body ``read()`` method.

        Parameters
        ---------
        object_name : str

        Returns
        -------
        stream
            file-like stream of object data
        """
        if not self.check_file_exists(object_name):
            raise IOError('Object "%s" does not exist' % object_name)
        s3_object = self.get_s3_object(object_name)
        return CloudStream(BytesIO(s3_object.get()['Body'].read()), sanitize_metadata(s3_object.metadata))

    def upload_file(self, file_name, cloud_name=None, permissions=DEFAULT_ACL):
        """Upload a file to S3.

        Parameters
        ----------
        file_name : str
            Absolute path of file to upload
        cloud_name : str, None
            Name of uploaded object. If None, use
            the full file name as the object name.
        permissions: ???
            S3 permissions

        Returns
        -------
        response : boto3 response
        """
        assert os.path.exists(file_name)
        if cloud_name is None:
            cloud_name = file_name
        s3_object = self.get_s3_object(cloud_name)
        return s3_object.upload_file(file_name, ExtraArgs = dict(ACL = permissions))

    def download_to_file(self, object_name, local_name):
        """Download S3 object to a file

        Parameters
        ----------
        object_name : str
        local_name : str
            Absolute path where the data will be downloaded on disk
        """
        assert self.check_file_exists(object_name)  # make sure object exists
        s3_object = self.get_s3_object(object_name)
        return s3_object.download_file(local_name)

    def upload_multipart(self, file_object, object_name, metadata, permissions=DEFAULT_ACL,
                         buffersize=MPU_CHUNKSIZE, verbose=True, **kwargs):
        """Multi-part upload for a python file-object.

        This automatically creates a multipart upload of an object.
        Useful for large objects that are loaded in memory. This avoids
        having to write the file to disk and then using ``upload_from_file``.

        Parameters
        ----------
        object_name : str
        file_object :
            file-like python object (e.g. StringIO, file, etc)
        buffersize  : int
            Byte size of the individual parts to create.
            Defaults to 100MB
        verbose     : bool
            verbosity flag of whether to print mpu information to stdout
        metadata  : optional
            Metadata to store along with MPU object
        permissions : str?
            permissions for this file
        """
        client=self.connection.meta.client

        mpu = client.create_multipart_upload(Bucket=self.bucket_name,
                                             ACL=permissions,
                                             Key=object_name,
                                             Metadata=metadata)

        # get size
        nbytes_total = get_fileobject_size(file_object)
        file_object.seek(0)

        # check if our buffersize is sensible
        if nbytes_total < buffersize:
            npotential = nbytes_total / float(MIN_MPU_SIZE)
            if npotential > 10:
                # 10MB sensible minimum for 50MB < x < 100MB (default)
                buffersize = MIN_MPU_SIZE * 2
            else:
                # this file is smaller than chunksize by a little
                buffersize = MIN_MPU_SIZE

        # figure out parts split
        nparts = int(np.floor(nbytes_total / float(buffersize)))
        last_part_offset = int(nbytes_total - nparts * buffersize)

        # make sure we can upload
        assert nbytes_total < MAX_MPU_SIZE  # 5TB
        assert buffersize >= MIN_MPU_SIZE  # 5MB
        assert nparts < MAX_MPU_PARTS  # 10,000
        assert nbytes_total > buffersize

        mpu_info = dict(Parts = [])
        if verbose:
            print('MPU: %0.02fMB file in %i parts' % (nbytes_total / (2. ** 20), nparts))

        for chunk_idx in range(nparts):
            part_number = chunk_idx + 1
            if part_number == nparts:
                buffersize += last_part_offset
            txt = (part_number, nparts, buffersize / (2. ** 20), nbytes_total / (2. ** 20))
            if verbose:
                print('Uploading %i/%i: %0.02fMB of %0.02fMB' % txt)

            data_chunk = file_object.read(buffersize)
            response = client.upload_part(Bucket = self.bucket_name,
                                          Key = object_name,
                                          UploadId = mpu['UploadId'],
                                          PartNumber = part_number,
                                          Body = data_chunk)

            # store the part info
            part_info = dict(PartNumber = part_number,
                             ETag = response['ETag'])
            mpu_info['Parts'].append(part_info)

        # finalize
        mpu_response = client.complete_multipart_upload(Bucket = self.bucket_name,
                                                        Key = object_name,
                                                        UploadId = mpu['UploadId'],
                                                        MultipartUpload = mpu_info)
        return mpu_response

    def copy(self, source, destination, source_bucket, destination_bucket, overwrite):
        source_bucket = self.get_bucket_name(source_bucket)
        dest_bucket = source_bucket if (destination_bucket is None) else destination_bucket
        dest_bucket = self.get_bucket_name(dest_bucket)

        assert self.check_file_exists(source, bucket_name = source_bucket)
        ob_new = self.get_s3_object(destination, bucket_name = dest_bucket)

        if self.check_file_exists(ob_new.key, bucket_name = dest_bucket):
            assert overwrite is True

        fpath = pathjoin(source_bucket, source)
        ob_new.copy_from(CopySource = fpath)
        return ob_new

    def move(self, source, destination, source_bucket, destination_bucket, overwrite):
        new_ob = self.copy(source, destination, source_bucket, destination_bucket, overwrite)
        old_ob = self.get_s3_object(source, bucket_name = source_bucket)
        old_ob.delete()
        return new_ob

    def list_directory(self, path, limit):
        """List the contents of a "directory"

        Parameters
        ----------
        path : str (default: "/")

        Returns
        -------
        matches : list
            The children of the path.
        """
        if has_real_magic(path):
            raise ValueError('Use ``ls()`` when using search patterns: "%s"' % path)

        if (path != '') and (path != '/'):
            path = remove_root(path)
        path = remove_trivial_magic(path)
        path = mk_aws_path(path)

        response = self.get_bucket().meta.client.list_objects(Bucket = self.bucket_name,
                                                              Delimiter = SEPARATOR,
                                                              Prefix = path,
                                                              MaxKeys = limit)
        object_names = []
        if 'CommonPrefixes' in response:
            # we got common paths
            object_list = [list(t.values()) for t in response['CommonPrefixes']]
            object_names += reduce(lambda x, y: x + y, object_list)
        if 'Contents' in response:
            # we got objects on the leaf nodes
            object_names += unquote_names([t['Key'] for t in response['Contents']])
        return [os.path.normpath(n) for n in object_names]

    def delete(self, object_name, recursive=False, delete=False):
        raise RuntimeError('Deleting on S3 backend is implemented by cottoncandy interface object')

    def get_object_metadata(self, object_name):
        """Get metadata associated with an object"""
        s3_object = self.get_s3_object(object_name)
        metadata = sanitize_metadata(s3_object.metadata)
        return metadata

    def get_object_size(self, object_name):
        """Get the size in bytes of an object"""
        s3_object = self.get_s3_object(object_name)
        size = s3_object.content_length
        return size



from __future__ import print_function
from pydrive.auth import GoogleAuth
from pydrive.drive import GoogleDrive
from pydrive.files import GoogleDriveFile, FileNotUploadedError, ApiRequestError
from .backend import CCBackEnd, FileNotFoundError, CloudStream
import six
import sys
import re

# Ipython autocomplete
try:
    # >=ipython-1.0
    from IPython import get_ipython
except ImportError:
    try:
        # support >=ipython-0.11, <ipython-1.0
        from IPython.core.ipapi import get as get_ipython
    except ImportError:
        # support <ipython-0.11
        try:
            from IPython.ipapi import get as get_ipython
        except ImportError:
            print('Not ipython')

if sys.version_info.major > 2:
    raw_input = input  # future compatibility


class GDriveClient(CCBackEnd):
    """
    Google Drive client based on PyDrive, which is based on google apis.
    To use, you need to enable gdrive APIs and make an OAuth2 id
    """

    @staticmethod
    def Authenticate(secrets, credentials):
        """

        Parameters
        ----------
        secrets : str
            path to client secrets json file
        credentials : str
            path to credentials file
        Returns
        -------
        authenticator : GoogleAuth
            google drive authentication object
        """
        authenticator = GoogleAuth()
        authenticator.settings = {
            'client_config_backend': 'file',
            'client_config_file':    secrets,
            'save_credentials':      False,
            'oauth_scope':           ['https://www.googleapis.com/auth/drive']
        }

        # get credentials
        authenticator.LoadCredentialsFile(credentials)
        if authenticator.credentials is None:  # no credentials
            response = raw_input('No credentials. Authenticate with local web browser? [y]/n > ')
            if response.lower() in ['y', 'yes'] or len(response) == 0:
                authenticator.LocalWebserverAuth()
            else:
                authenticator.CommandLineAuth()
        elif authenticator.access_token_expired:  # stale credentials
            authenticator.Refresh()
        else:
            authenticator.Authorize()

        authenticator.SaveCredentialsFile(credentials)  # update/save credentials

        return authenticator

    def __init__(self, secrets='client_secrets.json', credentials='gdrive-credentials.txt'):
        """

        Parameters
        ----------
        secrets : str
            path to client secrets json file
        credentials : str
            path to credentials file

        Returns
        -------

        """

        super(GDriveClient, self).__init__()

        # fields
        self.drive = None
        self.current_directory_object = None
        self.dir = None
        self.service = None
        self.rootID = None

        ### Authentication
        authenticator = GDriveClient.Authenticate(secrets, credentials)

        ### Get actual drive client
        self.initialise_drive(authenticator)

        # if running in ipython, hook in autocomplete
        try:
            self.hook_ipython_completer()
        except Exception as e:
            print(e.__str__())

    def initialise_drive(self, authenticator):
        """

        Parameters
        ----------
        authenticator : GoogleAuth
            authenticated GoogleAuth object

        Returns
        -------

        """
        self.drive = GoogleDrive(authenticator)
        self.current_directory_object = self.get_file_by_ID('root')  # cmd-like movement in google drive structure
        self.rootID = self.current_directory_id  # remember root ID
        self.dir = '/'
        self.list_objects(True)  # this call inits self.drive.auth.service, which exposes lower-level api calls
        self.service = self.drive.auth.service  # less typing XD

    @property
    def current_directory_id(self):
        """ID of current directory
        Returns
        -------
        id : str
            Google UUID
        """
        if self.current_directory_object.metadata['id'] == self.rootID:
            return 'root'
        else:
            return self.current_directory_object.metadata['id']

    #### Navigation commands

    @property
    def pwd(self):
        """Prints current working directory

        Returns
        -------
        None
        """
        print(self.dir)
        return None

    def list_directory(self, path, limit):
        if path is not None:
            path = re.sub('^\./', '', path)
        current_directory = self.current_directory_object
        if path is not None and len(path) > 0:
            if not self.cd(path):
                return
        list = self.list_objects(names_only = True)
        if path is not None:
            self.current_directory_object = current_directory
            self.rebuild_current_path()
        return list

    def ls(self, directory=None):
        """Prints contents of a folder

        Parameters
        ----------
        directory : str
            folder to list, if None defaults to pwd

        Returns
        -------
        None
        """
        if directory is not None:
            directory = re.sub('^\./', '', directory)
        current_directory = self.current_directory_object
        if directory is not None and len(directory) > 0:
            if not self.cd(directory):
                return
        list = self.list_objects()
        for item in list:
            print('{} {}'.format('d' if item['mimeType'] == 'application/vnd.google-apps.folder' else ' ', item['title']))
        if directory is not None:
            self.current_directory_object = current_directory
            self.rebuild_current_path()

    def cd(self, directory=None, make_if_not_exist=False, isID=False):
        """Changes directory

        Parameters
        ----------
        directory : str
            path
        make_if_not_exist : bool
            make the directory if it doesn't exist?
        isID : bool
            is `directory` a Google UUID?
        Returns
        -------
        : bool
            success of cd operation
        """
        if directory is None or directory == '.':
            print(self.dir)
            return True

        if isID:
            try:
                self.current_directory_object = self.get_file_by_ID(directory)
                self.rebuild_current_path()
                return True
            except FileNotFoundError:
                return False

        directory = re.sub('^\./', '', directory)

        remainder = None
        if '/' in directory:  # not a simple change, so we just process the first one only each turn
            directory = re.sub('/$', '', directory)
            if '/' in directory:  # if it wasn't just a ending slash
                dirs = re.split('/', directory, 1)
                directory = dirs[0]
                remainder = dirs[1]

        if directory == '..':  # move up one
            if self.current_directory_object['parents'][0]['isRoot']:
                self.current_directory_object = self.get_file_by_ID('root')
            else:
                self.current_directory_object = self.get_file_by_ID(self.current_directory_object['parents'][0]['id'])
        elif directory in ['/', '']:  # change to root
            self.current_directory_object = self.get_file_by_ID('root')
        else:  # move down one
            directories = self.drive.ListFile({'q': 'title="{}" and "{}" in parents and mimeType = \'application/vnd.google-apps.folder\' and trashed = false'.format(directory, self.current_directory_id)}).GetList()
            if len(directories) < 1:
                if not make_if_not_exist:
                    print('No such directory: {}'.format(directory))
                    return False
                else:
                    directories.append(self.get_file_by_ID(self.mkdir(directory)))
            if len(directories) > 1:
                print('For some reason there\'s more than one result')
                for item in directories:
                    print('{}\t{}'.format(item['id'], item['title']))
                return False
            self.current_directory_object = directories[0]

        self.rebuild_current_path()
        # recursively cd to the innermost folder
        if remainder is not None:
            return self.cd(remainder)
        return True

    def mkdir(self, folder_name):
        """

        Parameters
        ----------
        folder_name : str
            name of folder to make

        Returns
        -------
        id : str
            Google UUID of new directory

        """
        # name formatting
        folder_name = re.sub('^\./', '', folder_name)

        # not current dir
        current_directory = None
        if '/' in folder_name:
            current_directory = self.current_directory_object
            tokens = re.split('/', folder_name)
            for i in range(len(tokens) - 1):
                if not self.cd(tokens[i], True):
                    return None
            folder_name = tokens[-1]

        directories = self.drive.ListFile({'q': 'title="{}" and "{}" in parents and mimeType = \'application/vnd.google-apps.folder\' and trashed=false'.format(folder_name, self.current_directory_id)}).GetList()
        if len(directories) > 0:
            print('Folder already exists')
            return None

        folder = self.drive.CreateFile({'title': folder_name, 'mimeType': 'application/vnd.google-apps.folder', 'parents': [{'id': self.current_directory_id}]})
        folder.Upload()

        if current_directory is not None:
            self.current_directory_object = current_directory
            self.rebuild_current_path()
        return folder.metadata['id']

    def move(self, file, destination, sb=None, db=None, o=None):
        """

        Parameters
        ----------
        file : str
            file to move
        destination : str
            destination folder

        Returns
        -------
        : bool
            success of move operation
        """

        # TODO: renaming in mv
        origin = re.match('.*/', file)
        destination = re.match('.*/', destination)
        if origin is destination:
            print('renaming using move is not supported yet. use the .rename method')
            return False
        if origin.group(0) == destination.group(0):
            print('renaming using move is not supported yet. use the .rename method')
            return False

        return self.update_metadata(file, {'parents': [{'id': self.get_ID_by_name(destination)}]})

    def rename(self, original_name, new_name):
        """

        Parameters
        ----------
        original_name : str
            file to rename
        new_name : str
            new name

        Returns
        -------
        : bool
            success of operation
        """

        # TODO: name validation
        if '/' in new_name:
            print('No slashes in name, please')
            return False

        return self.update_metadata(original_name, {'title': new_name})

    def copy(self, original_name, copy_name, sb=None, db=None, o=None):
        """

        Parameters
        ----------
        original_name : str
            name of origin
        copy_name : str
            name/path of copy

        Returns
        -------

        """
        # TODO: directory copying and deep/shallowness
        try:
            original = self.get_file_by_name(original_name)
            self.upload_stream(copy_name, original.content)
            copyID = self.get_ID_by_name(copy_name)
            if 'properties' in original.metadata.keys() and len(original.metadata['properties'].keys()) > 0:
                for key in original.metadata['properties'].keys():
                    if not self.insert_property(copyID, key, original.metadata['properties'][key]):
                        print('Property copy for {} failed'.format(key))
            return True
        except Exception as e:
            print (e.__str__())
            return False

    def delete(self, file_name, recursive=False, delete=False):
        """

        Parameters
        ----------
        file_name : str
            name of file/folder to delete
        recursive : bool
            recursively delete a folder?
        delete : bool
            hard delete files?

        Returns
        -------
        : bool
            success of operation
        """
        try:
            fileID = self.get_ID_by_name(file_name)
            f = self.drive.CreateFile({'id': fileID})
            f.FetchMetadata()
            if f.metadata['mimeType'] == 'application/vnd.google-apps.folder' and not recursive:
                print('Is folder and recursive delete is not selected')
                return False
            if delete:
                f.Delete()
            else:
                f.Trash()  # trashed files still exist so the association count remains the same
            return True
        except FileNotFoundError:
            print('File not found')
            return False

    #### File IO functions

    def upload_file(self, file_name, cloud_name=None, permissions=None):
        """Uploads from file on disk

        Parameters
        ----------
        file_name : str
            path to file on disk to upload
        cloud_name : str
            path/name for cloud file

        Returns
        -------
        : bool
            success of operation
        """
        if cloud_name is None:
            cloud_name = file_name

        if not isinstance(file_name, six.string_types):
            # assume this is a file_name-like object
            self.upload_stream(file_name, cloud_name)

        # cloud_name formatting
        cloud_name = re.sub('^\./', '', cloud_name)

        # not current dir
        current_directory = None
        if '/' in cloud_name:
            current_directory = self.current_directory_object
            tokens = re.split('/', cloud_name)
            for i in range(len(tokens) - 1):
                if not self.cd(tokens[i], True):
                    return False
            cloud_name = tokens[-1]

        # metadata
        metadata = {}
        metadata['title'] = cloud_name
        metadata['parents'] = [{'id': self.current_directory_id}]

        # try to upload the file_name
        newfile = self.drive.CreateFile(metadata)
        newfile.Upload()
        try:
            newfile.SetContentFile(file_name)
            newfile.Upload()
        except Exception as e:
            print('Error uploading file_name:\n{}'.format(e.__str__()))
            newfile.delete()
            return False

        if current_directory is not None:
            self.current_directory_object = current_directory
            self.rebuild_current_path()

        return True

    def upload_stream(self, stream, name, properties=None, permissions=None):
        """Upload a stream with a .read() method

        Parameters
        ----------
        stream : stream
            stream to upload
        name : str
            name to use for cloud file
        properties : dict
            custom metadata

        Returns
        -------
        : bool
            success of operation
        """
        # cloudName formatting
        name = re.sub('^./', '', name)

        # not current dir
        current_directory = None
        if '/' in name:
            current_directory = self.current_directory_object
            tokens = re.split('/', name)
            for i in range(len(tokens) - 1):
                if not self.cd(tokens[i], True):
                    return False
            name = tokens[-1]

        # metadata
        metadata = {}
        metadata['title'] = name
        metadata['parents'] = [{'id': self.current_directory_id}]

        newfile = self.drive.CreateFile(metadata)
        newfile.Upload()
        try:
            newfile.content = stream
            newfile.Upload()
        except Exception as e:
            print('Error uploading stream:\n{}'.format(e.__str__()))
            newfile.delete()
            return False

        if properties is not None:
            for key in properties.keys():
                if not self.insert_property(newfile.metadata['id'], key, properties[key]):
                    print('Property insertion for {} failed'.format(key))

        if current_directory is not None:
            self.current_directory_object = current_directory
            self.rebuild_current_path()

        return True

    def upload_multipart(self, stream, cloud_name, properties=None, permissions=None, buffersize = -1, verbose = False):
        return self.upload_stream(stream, cloud_name, properties, permissions)


    def download_to_file(self, drive_file, local_file=None):
        """Download a file to disk

        Parameters
        ----------
        drive_file : str
            name of file to download
        local_file : str
            name to use on disk, if None, uses same name

        Returns
        -------
        : bool
            success of operation
        """

        try:
            f = self.drive.CreateFile({'id': self.get_ID_by_name(drive_file)})
        except FileNotFoundError:
            print('File not found')
            return False

        f.FetchMetadata()
        if local_file is None:
            local_file = f.metadata['title']
        f.GetContentFile(local_file)

        return True

    def download_stream(self, drive_file):
        """Downloads a file to memory

        Parameters
        ----------
        drive_file : str
            name of file to download

        Returns
        -------
        : CloudStream
            object in memory of downloaded data

        Raises
        ------
        FileNotFoundError
        """

        try:
            f = self.drive.CreateFile({'id': self.get_ID_by_name(drive_file)})
        except FileNotFoundError:
            print('File not found')
            raise FileNotFoundError

        f.FetchMetadata()
        properties = self.get_file_properties(f.metadata['id'])
        if 'key' in properties:
            properties['key'] = self.get_encryption_key(f.metadata['id'])

        if f.content is None:
            f.FetchContent()
        return CloudStream(f.content, properties)

    #### Misc helper functions

    def check_ID_exists(self, id):
        """

        Parameters
        ----------
        id : str
            Google UUID

        Returns
        -------
        : bool
            existence of ID on drive
        """
        metadata = {'id': id}
        item = GoogleDriveFile(self.drive.auth, metadata, uploaded = True)
        try:
            item.FetchMetadata()
            return True
        except (FileNotUploadedError, ApiRequestError):
            return False
        print('This line should not be executed.')

    def check_file_exists(self, file_name, bn):
        """

        Parameters
        ----------
        file_name : str
            path on cloud to check

        Returns
        -------
        : bool
            existence of file
        """
        try:
            self.get_ID_by_name(file_name)
            return True
        except FileNotFoundError:
            return False
        print('This line should not be executed.')


    def rebuild_current_path(self):
        """After a cd, rebuilds the current working directory string
        """
        if self.current_directory_id == 'root':
            self.dir = '/'
            return

        self.dir = ''
        this_directory = self.current_directory_object
        self.dir = '/' + this_directory['title'] + self.dir
        while (not this_directory['parents'][0]['isRoot']):
            this_directory = self.get_file_by_ID(self.current_directory_object['parents'][0]['id'])
            self.dir = '/' + this_directory['title'] + self.dir
        self.dir = str(self.dir)

    def list_objects(self, names_only=False, trashed=False):
        """Gets a list of all files in current directory

        Parameters
        ----------
        names_only : bool
            Returns names only?
        trashed : bool
            Look in trash folder instead?

        Returns
        -------
        : list<str> | list<GoogleDriveFile>
            List of files
        """
        if names_only:
            files = self.drive.ListFile({'q': "'{}' in parents and trashed={}".format(self.current_directory_id, str(trashed).lower())}).GetList()
            out = []
            for f in files:
                out.append(f['title'])
            return out
        return self.drive.ListFile({'q': "'{}' in parents and trashed={}".format(self.current_directory_id, str(trashed).lower())}).GetList()

    def get_file_by_name(self, name):
        """Gets the GoogleDriveFile for a file
        Parameters
        ----------
        name : str
            name of file to get

        Returns
        -------
        item : GoogleDriveFile
            object for the file
        """
        return self.get_file_by_ID(self.get_ID_by_name(name))

    def get_file_by_ID(self, id):
        """Gets the GoogleDriveFile for a file

        Parameters
        ----------
        id : str
            Google UUID

        Returns
        -------
        item : GoogleDriveFile
            object for the file
        """
        metadata = {'id': id}
        item = GoogleDriveFile(self.drive.auth, metadata, uploaded = True)
        try:
            item.FetchMetadata()
        except (FileNotUploadedError, ApiRequestError):
            print('File does not exist')
            return None
        return item

    def get_ID_by_name(self, file_name):
        """Gets the Google UUID for a file

        Parameters
        ----------
        file_name : str
            name of file to get

        Returns
        -------
        id : str
            Google UUID
        """

        drive_file = re.sub('^\./', '', file_name)

        if file_name == '/':
            return 'root'

        # not current dir
        current_directory = None
        if '/' in drive_file:
            current_directory = self.current_directory_object
            tokens = re.split('/', drive_file)
            if ('' in tokens):
                tokens.remove('')
            for i in range(len(tokens) - 1):
                if not self.cd(tokens[i]):
                    raise FileNotFoundError
            drive_file = tokens[-1]

        items = self.list_objects()
        item = None
        for i in items:
            if i['title'] == drive_file:
                item = i
                break

        if current_directory is not None:
            self.current_directory_object = current_directory
            self.rebuild_current_path()

        if item is not None:
            return item['id']
        else:
            raise FileNotFoundError

    def insert_property(self, id, key, value, visibility='PUBLIC'):
        """Adds a custom property to a file

        Parameters
        ----------
        id : str
            Google UUID of file
        key : str
            name of the custom property
        value : str
            value of the custom property
        visibility : 'PUBLIC'|'PRIVATE'
            visibility of the property

        Returns
        -------
        : bool
            operation success

        """

        if visibility not in ['PUBLIC', 'PRIVATE']:
            raise ValueError('Bad visibility value')

        if key == 'key':
            return self.store_encryption_key(id, value)

        body = {'key': key, 'value': value, visibility: visibility}
        # print('Adding {}: {}'.format(key, value))

        try:
            self.service.properties().insert(fileId = id, body = body).execute()
            return True
        except Exception as e:
            print('Error: {}'.format(e.__str__()))
            return False

    def get_file_properties(self, id):
        """Gets the properties for a file

        Parameters
        ----------
        id : str
            Google UUID of file

        Returns
        -------
        properties : dict
            custom metadata
        """
        try:
            f = self.get_file_by_ID(id)
            if f is None:
                print('File does not exist')
            properties = {}
            f.FetchMetadata()
            if 'properties' in f.metadata.keys():
                for p in f.metadata['properties']:
                    properties[p['key']] = p['value']
            return properties
        except Exception as e:
            print('Error: {}'.format(e.__str__()))
            return None

    def store_encryption_key(self, fileID, key64, chunk_size=96):
        """Stores a base64-encoded encryption key

        Parameters
        ----------
        fileID : str
            file UUID to store this to
        key64 : str
            base-64 encoded key
        chunk_size : int
            size in chars to store the key in (there are limits to property lengths)

        Returns
        -------

        """
        if chunk_size > 114:
            print('Chunk size set to 114 because of limitations of metadata size')
            chunk_size = 114
        nChuncks = len(key64) / chunk_size + (len(key64) % chunk_size > 0)
        self.insert_property(fileID, 'keyChunks', str(nChuncks))
        self.insert_property(fileID, 'key', 'in chunks')
        for i in range(nChuncks):
            start = i * chunk_size
            end = (i + 1) * chunk_size
            end = end if end < len(key64) else len(key64)
            if not self.insert_property(fileID, 'keyChunk{}'.format(i), key64[start:end]):
                print('Key upload failed')
                return

    def get_encryption_key(self, fileID):
        """Gets the encryption key from the properties for a file

        Parameters
        ----------
        fileID : str
            Google UUID

        Returns
        -------
        key64 : str
            base-64 encoded encryption key
        """
        key64 = ''
        properties = self.get_file_properties(fileID)
        if 'keyChunks' not in properties:
            print('No stored key')
            return None
        else:
            nChunks = int(properties['keyChunks'])
            for i in range(nChunks):
                key64 += properties['keyChunk{}'.format(i)]
            return key64

    def update_metadata(self, file_name, metadata):
        """Updates the custom properties for a file

        Parameters
        ----------
        file_name : str
            Name of file
        metadata : dict
            new metadata

        Returns
        -------
        : bool
            operation success
        """
        try:
            drive_file = self.service.files().get(fileId = self.get_ID_by_name(file_name)).execute()
            for key in metadata.keys():
                drive_file[key] = metadata[key]
            self.service.files().update(fileId = self.get_ID_by_name(file_name), body = drive_file).execute()
            return True
        except FileNotFoundError:
            print('File not found')
            return False
        except Exception as e:
            print('Update failed')
            print(e.__str__())
            return False

    @property
    def size(self):
        files = self.drive.ListFile({'q': "trashed=false"}).GetList()
        sizes = [f.metadata['size'] for f in files]
        return sum(sizes)

    def completer(self, context, event):
        """Ipython autocomplete hook

        Parameters
        ----------
        context : ???
            ???
        event : ???
            eventargs

        Returns
        -------
        : list<str>
            autocompelte options
        """
        # print(event.line)
        path = re.sub('(?:.*\=)?(.+?)(\.(ls|cd|Download.?|rm|mv|rename|CheckFileExists|CheckIDExists))\([\'"]', '', event.line)  # remove the user command, leaving only the path
        path = re.sub('.+[\'"]\s?,\s?[\'"]', '', path)  # for cases like mv where there may be a second argument, removes the first arg if complete
        # path = re.sub('^\./', '', path)

        if '/' in path:
            # print('/ foound')
            directory = re.match('^.*/', path).group(0)
            path = re.search('[^/]+$', path)
            if path is not None:
                path = path.group(0)
            # print('\n{}\n{}\n'.format(directory, path))

            current_directory = self.current_directory_object
            if directory is not None:
                if not self.cd(directory):  # if the directory doesn't actually exist, return empty list
                    return []
            items = self.list_objects(True)
            if directory is not None:
                self.current_directory_object = current_directory
                self.rebuild_current_path()
            out = []
            for item in items:
                out.append(directory + item)
            # print(out)
            return out
        else:
            items = self.list_objects(True)

            if path is None:
                return items
            out = []
            for item in items:
                # if re.match(path, item) is not None:	# Ipython takes care of filtering the matches
                out.append(item)
            return out

    def hook_ipython_completer(self):
        """Hooks the autocompleter into ipython
        """
        ipython = get_ipython()
        if ipython is None:
            return
        ipython.set_hook('complete_command', self.completer, re_key = r'(?:.*\=)?(.+?)(\.(ls|cd|Download|rm|DownloadRawArray|DownloadStream|mv|rename|CheckFileExists|CheckIDExists))\([\'"].*')

    def get_object_metadata(self, object_name):
        """Get metadata associated with an object"""
        f = self.drive.CreateFile({'id': self.get_ID_by_name(object_name)})

        f.FetchMetadata()
        properties = self.get_file_properties(f.metadata['id'])
        if 'key' in properties:
            properties['key'] = self.get_encryption_key(f.metadata['id'])

        return properties

    def get_object_size(self, object_name):
        """Get the size in bytes of an object"""
        f = self.drive.CreateFile({'id': self.get_ID_by_name(object_name)})
        f.FetchMetadata()
        size = f.metadata['size']
        return size
